from typing import Dict, List, Type, TypeVar, Generic

from ms_deisotope.peak_set import IonMobilityProfileDeconvolutedPeakSolution
from ms_deisotope.data_source.scan.mobility_frame import IonMobilityFrame, IonMobilitySourceRandomAccessFrameSource
from ms_deisotope.feature_map.feature_map import DeconvolutedLCMSFeatureForest, IonMobilityProfileDeconvolutedLCMSFeatureForest
from ms_deisotope.data_source.metadata.scan_traits import IsolationWindow


FeatureForestType = TypeVar(
    'FeatureForestType', bound=DeconvolutedLCMSFeatureForest)


class IsolationWindowDemultiplexerBase(Generic[FeatureForestType]):
    forest_type: Type[FeatureForestType]
    error_tolerance: float
    forest_for_isolation_window: Dict[IsolationWindow, FeatureForestType]

    def forest_for(self, isolation_window: IsolationWindow) -> FeatureForestType:
        """Get the :class:`~.DeconvolutedLCMSFeatureForest` for a specific isolation window,
        or if one does not exist, create it.

        Parameters
        ----------
        isolation_window : :class:`~.IsolationWindow`
            The isolation window to select with

        Returns
        -------
        :class:`~.DeconvolutedLCMSFeatureForest`
        """
        if isolation_window is None:
            raise ValueError("An isolation window None!")
        try:
            forest = self.forest_for_isolation_window[isolation_window]
        except KeyError:
            forest = self.forest_type(error_tolerance=self.error_tolerance)
            self.forest_for_isolation_window[isolation_window] = forest
        return forest

    def all_forests_overlapping_window(self, isolation_window: IsolationWindow) -> List[FeatureForestType]:
        """Get all :class:`~.DeconvolutedLCMSFeatureForest` that are from isolation windows
        which overlap with `isolation_window`.

        The result list is ordered by isolation window lower bound.

        Parameters
        ----------
        isolation_window : :class:`~.IsolationWindow`
            The isolation window to query for overlaps with

        Returns
        -------
        list of (:class:`~.IsolationWindow`, :class:`~.DeconvolutedLCMSFeatureForest`) pairs.
            All feature
        """
        forests = []
        lb = isolation_window.lower_bound
        ub = isolation_window.upper_bound
        for window, forest in self.forest_for_isolation_window.items():
            if window.spans(lb) or window.spans(ub):
                forests.append((window, forest))
        if len(forests) > 1:
            forests.sort(key=lambda x: x[0].lower_bound)
        return forests


class IsolationWindowDemultiplexingDeconvolutedLCMSFeatureForestCollection(IsolationWindowDemultiplexerBase[DeconvolutedLCMSFeatureForest]):
    """Aggregate MS2 features per isolation window over the course of a
    deconvoluted LC-MS/MS run.

    Attributes
    ----------
    reader : ProcessedMzMLDeserializer
        A reader over a deconvoluted LC-MS/MS run
    minimum_mass : float
        The mass below which MS2 peaks will be ignored
    minimum_intensity : float
        The intensity below which MS2 peaks will be ignored
    error_tolerance : float
        The mass error tolerance (PPM) when aggregating peaks into features
    delta_rt : float
        The maximum gap size to tolerate in the time dimension

    """
    forest_type = DeconvolutedLCMSFeatureForest

    minimum_mass: float
    minimum_intensity: float


    def __init__(self, reader, minimum_mass=80.0, minimum_intensity=10.0, error_tolerance=1e-5, delta_rt=1.0):
        self.reader = reader
        self.error_tolerance = error_tolerance
        self.minimum_mass = minimum_mass
        self.minimum_intensity = minimum_intensity
        self.delta_rt = delta_rt
        self.forest_for_isolation_window = dict()

    def handle_scan(self, scan):
        """Add `scan`'s peaks to the appropriate feature forest by its isolation window.

        Parameters
        ----------
        scan : :class:`~.ScanBase`
            The deconvoluted scan to process
        """
        isolation_window = scan.isolation_window
        forest = self.forest_for(isolation_window)
        time = scan.scan_time
        for peak in scan.deconvoluted_peak_set:
            if peak.neutral_mass >= self.minimum_mass and peak.intensity >= self.minimum_intensity:
                forest.handle_peak(peak, time)

    def complete(self):
        """Apply any last post-processing to the feature forests, smoothing
        overlapping features and breaking gaps.
        """
        for key, forest in self.forest_for_isolation_window.items():
            forest.smooth_overlaps()
            forest.split_sparse(self.delta_rt)

    def aggregate_peaks(self):
        """Run the aggregation process to completion, and post-process the results,
        producing a fully-fledged set of :class:`~.DeconvolutedLCMSFeatureForest`
        """
        self.reader.reset()
        self.reader.make_iterator(grouped=False)
        n = len(self.reader)
        for i, scan in enumerate(self.reader):
            if i % 1000 == 0 and i:
                print("Aggregated %d/%d Scans (%0.2f%%)" %
                      (i, n, i * 100.0 / n))
            if scan.ms_level > 1:
                self.handle_scan(scan)
        self.complete()

    @classmethod
    def demultiplex(cls, reader, minimum_mass=80.0, minimum_intensity=10.0, error_tolerance=1e-5, delta_rt=1.0):
        """Aggregate MS2 features per isolation window over the course of a deconvoluted LC-MS/MS run.

        Parameters
        ----------
        reader : ProcessedMzMLDeserializer
            A reader over a deconvoluted LC-MS/MS run
        minimum_mass : float
            The mass below which MS2 peaks will be ignored
        minimum_intensity : float
            The intensity below which MS2 peaks will be ignored
        error_tolerance : float
            The mass error tolerance (PPM) when aggregating peaks into features
        delta_rt : float
            The maximum gap size to tolerate in the time dimension

        Returns
        -------
        IsolationWindowDemultiplexingDeconvolutedLCMSFeatureForestCollection
        """
        task = cls(reader, minimum_mass, minimum_intensity,
                   error_tolerance, delta_rt)
        task.aggregate_peaks()
        return task


demultiplex_ms2_features = IsolationWindowDemultiplexingDeconvolutedLCMSFeatureForestCollection.demultiplex


class DemultiplexingIonMobilityProfileDeconvolutedLCMSFeatureForestCollection(IsolationWindowDemultiplexerBase[IonMobilityProfileDeconvolutedLCMSFeatureForest]):
    forest_type = IonMobilityProfileDeconvolutedLCMSFeatureForest

    reader: IonMobilitySourceRandomAccessFrameSource
    minimum_mass: float
    delta_rt: float
    precursor_forest: IonMobilityProfileDeconvolutedLCMSFeatureForest

    def __init__(self, reader: IonMobilitySourceRandomAccessFrameSource, minimum_mass: float=80.0, error_tolerance: float=1e-5, delta_rt: float=1.0):
        self.reader = reader
        self.error_tolerance = error_tolerance

        self.minimum_mass = minimum_mass
        self.delta_rt = delta_rt
        self.forest_for_isolation_window = dict()

        self.precursor_forest = self.forest_type(error_tolerance=self.error_tolerance)

    def handle_ms1_frame(self, frame: IonMobilityFrame):
        time = frame.time
        for feature in frame.deconvoluted_features:
            if feature.neutral_mass < self.minimum_mass:
                continue
            peak = IonMobilityProfileDeconvolutedPeakSolution.from_feature(feature)
            self.precursor_forest.handle_peak(peak, time)

    def handle_msn_frame(self, frame: IonMobilityFrame):
        isolation_window = frame.isolation_window
        forests = self.all_forests_overlapping_window(isolation_window)
        time = frame.time
        for feature in frame.deconvoluted_features:
            if feature.neutral_mass < self.minimum_mass:
                continue
            peak = IonMobilityProfileDeconvolutedPeakSolution.from_feature(feature)
            for forest in forests:
                forest.handle_peak(peak, time)

    def complete(self):
        """Apply any last post-processing to the feature forests, smoothing
        overlapping features and breaking gaps.
        """
        self.precursor_forest.smooth_overlaps().split_sparse(self.delta_rt)
        for _key, forest in self.forest_for_isolation_window.items():
            forest.smooth_overlaps()
            forest.split_sparse(self.delta_rt)



""" Waters
    MassLynx Python SDK
"""

from enum import IntEnum

class MassLynxBaseType(IntEnum):
    SCAN = 1
    INFO = 2
    CHROM = 3
    ANALOG = 4
    LOCKMASS = 5
    CENTRIOD = 6
    DDA = 7
    MSE = 8
    UNINITALISED = 99

class MassLynxIonMode(IntEnum):
    EI_POS = 100
    EI_NEG = EI_POS + 1
    CI_POS = EI_POS + 2
    CI_NEG = EI_POS + 3
    FB_POS = EI_POS + 4
    FB_NEG = EI_POS + 5
    TS_POS = EI_POS + 6
    TS_NEG = EI_POS + 7
    ES_POS = EI_POS + 8
    ES_NEG = EI_POS + 9
    AI_POS = EI_POS + 10
    AI_NEG = EI_POS + 11
    LD_POS = EI_POS + 12
    LD_NEG = EI_POS + 13
    UNINITIALISED = EI_POS + 99

class MassLynxFunctionType(IntEnum):
    MS = 200
    SIR = 1 + MS
    DLY = 2 + MS
    CAT = 3 + MS
    OFF = 4 + MS
    PAR = 5 + MS
    DAU = 6 + MS
    NL = 7 + MS
    NG = 8 + MS
    MRM = 9 + MS
    Q1F = 10 + MS
    MS2 = 11 + MS
    DAD = 12 + MS
    TOF = 13 + MS
    PSD = 14 + MS
    TOFS = 15 + MS
    TOFD = 16 + MS
    MTOF = 17 + MS
    TOFM = 18 + MS
    TOFP = 19 + MS
    ASVS = 20 + MS
    ASMS = 21 + MS
    ASVSIR = 22 + MS
    ASMSIR = 23 + MS
    QUADD = 24 + MS
    ASBE = 25 + MS
    ASB2E = 26 + MS
    ASCNL = 27 + MS
    ASMIKES = 28 + MS
    ASMRM = 29 + MS
    ASNRMS = 30 + MS
    ASMRMQ = 31 + MS
    UNINITIALISED = MS + 99

class MassLynxHeaderItem(IntEnum):
    VERSION = 300
    ACQUIRED_NAME = 1 + VERSION
    ACQUIRED_DATE = 2 + VERSION
    ACQUIRED_TIME = 3 + VERSION
    JOB_CODE = 4 + VERSION
    TASK_CODE = 5 + VERSION
    USER_NAME = 6 + VERSION
    INSTRUMENT = 7 + VERSION
    CONDITIONS = 8 + VERSION
    LAB_NAME = 9 + VERSION
    SAMPLE_DESCRIPTION = 10 + VERSION
    SOLVENT_DELAY = 11 + VERSION
    SUBMITTER = 12 + VERSION
    SAMPLE_ID = 13 + VERSION
    BOTTLE_NUMBER = 14 + VERSION
    ANALOG_CH1_OFFSET = 15 + VERSION
    ANALOG_CH2_OFFSET = 16 + VERSION
    ANALOG_CH3_OFFSET = 17 + VERSION
    ANALOG_CH4_OFFSET = 18 + VERSION
    CAL_MS1_STATIC = 19 + VERSION
    CAL_MS2_STATIC = 20 + VERSION
    CAL_MS1_STATIC_PARAMS = 21 + VERSION
    CAL_MS1_DYNAMIC_PARAMS = 22 + VERSION
    CAL_MS2_STATIC_PARAMS = 23 + VERSION
    CAL_MS2_DYNAMIC_PARAMS = 24 + VERSION
    CAL_MS1_FAST_PARAMS = 25 + VERSION
    CAL_MS2_FAST_PARAMS = 26 + VERSION
    CAL_TIME = 27 + VERSION
    CAL_DATE = 28 + VERSION
    CAL_TEMPERATURE = 29 + VERSION
    INLET_METHOD = 30 + VERSION
    SPARE1 = 31 + VERSION
    SPARE2 = 32 + VERSION
    SPARE3 = 33 + VERSION
    SPARE4 = 34 + VERSION
    SPARE5 = 35 + VERSION


class MassLynxScanItem(IntEnum):
    LINEAR_DETECTOR_VOLTAGE = 401
    LINEAR_SENSITIVITY = LINEAR_DETECTOR_VOLTAGE + 1
    REFLECTRON_LENS_VOLTAGE = LINEAR_DETECTOR_VOLTAGE + 2
    REFLECTRON_DETECTOR_VOLTAGE = LINEAR_DETECTOR_VOLTAGE + 3
    REFLECTRON_SENSITIVITY = LINEAR_DETECTOR_VOLTAGE + 4
    LASER_REPETITION_RATE = LINEAR_DETECTOR_VOLTAGE + 5
    COURSE_LASER_CONTROL = LINEAR_DETECTOR_VOLTAGE + 6
    FINE_LASER_CONTROL = LINEAR_DETECTOR_VOLTAGE + 7
    LASERAIM_XPOS = LINEAR_DETECTOR_VOLTAGE + 8
    LASERAIM_YPOS = LINEAR_DETECTOR_VOLTAGE + 9
    NUM_SHOTS_SUMMED = LINEAR_DETECTOR_VOLTAGE + 10
    NUM_SHOTS_PERFORMED = LINEAR_DETECTOR_VOLTAGE + 11
    SEGMENT_NUMBER = LINEAR_DETECTOR_VOLTAGE + 12
    LCMP_TFM_WELL = LINEAR_DETECTOR_VOLTAGE + 13
    SEGMENT_TYPE = LINEAR_DETECTOR_VOLTAGE + 14
    SOURCE_REGION1 = LINEAR_DETECTOR_VOLTAGE + 15
    SOURCE_REGION2 = LINEAR_DETECTOR_VOLTAGE + 16
    REFLECTRON_FIELD_LENGTH = LINEAR_DETECTOR_VOLTAGE + 17
    REFLECTRON_LENGTH = LINEAR_DETECTOR_VOLTAGE + 18
    REFLECTRON_VOLT = LINEAR_DETECTOR_VOLTAGE + 19
    SAMPLE_PLATE_VOLT = LINEAR_DETECTOR_VOLTAGE + 20
    REFLECTRON_FIELD_LENGTH_ALT = LINEAR_DETECTOR_VOLTAGE + 21
    REFLECTRON_LENGTH_ALT = LINEAR_DETECTOR_VOLTAGE + 22
    PSD_STEP_MAJOR = LINEAR_DETECTOR_VOLTAGE + 23
    PSD_STEP_MINOR = LINEAR_DETECTOR_VOLTAGE + 24
    PSD_FACTOR_1 = LINEAR_DETECTOR_VOLTAGE + 25
    NEEDLE = LINEAR_DETECTOR_VOLTAGE + 49
    COUNTER_ELECTRODE_VOLTAGE = LINEAR_DETECTOR_VOLTAGE + 50
    SAMPLING_CONE_VOLTAGE = LINEAR_DETECTOR_VOLTAGE + 51
    SKIMMER_LENS = LINEAR_DETECTOR_VOLTAGE + 52
    SKIMMER = LINEAR_DETECTOR_VOLTAGE + 53
    PROBE_TEMPERATURE = LINEAR_DETECTOR_VOLTAGE + 54
    SOURCE_TEMPERATURE = LINEAR_DETECTOR_VOLTAGE + 55
    RF_VOLTAGE = LINEAR_DETECTOR_VOLTAGE + 56
    SOURCE_APERTURE = LINEAR_DETECTOR_VOLTAGE + 57
    SOURCE_CODE = LINEAR_DETECTOR_VOLTAGE + 58
    LM_RESOLUTION = LINEAR_DETECTOR_VOLTAGE + 59
    HM_RESOLUTION = LINEAR_DETECTOR_VOLTAGE + 60
    COLLISION_ENERGY = LINEAR_DETECTOR_VOLTAGE + 61
    ION_ENERGY = LINEAR_DETECTOR_VOLTAGE + 62
    MULTIPLIER1 = LINEAR_DETECTOR_VOLTAGE + 63
    MULTIPLIER2 = LINEAR_DETECTOR_VOLTAGE + 64
    TRANSPORTDC = LINEAR_DETECTOR_VOLTAGE + 65
    TOF_APERTURE = LINEAR_DETECTOR_VOLTAGE + 66
    ACC_VOLTAGE = LINEAR_DETECTOR_VOLTAGE + 67
    STEERING = LINEAR_DETECTOR_VOLTAGE + 68
    FOCUS = LINEAR_DETECTOR_VOLTAGE + 69
    ENTRANCE = LINEAR_DETECTOR_VOLTAGE + 70
    GUARD = LINEAR_DETECTOR_VOLTAGE + 71
    TOF = LINEAR_DETECTOR_VOLTAGE + 72
    REFLECTRON = LINEAR_DETECTOR_VOLTAGE + 73
    COLLISION_RF = LINEAR_DETECTOR_VOLTAGE + 74
    TRANSPORT_RF = LINEAR_DETECTOR_VOLTAGE + 75
    SET_MASS = LINEAR_DETECTOR_VOLTAGE + 76,
    COLLISION_ENERGY2 = LINEAR_DETECTOR_VOLTAGE + 77
    SET_MASS_CALL_SUPPORTED = LINEAR_DETECTOR_VOLTAGE + 78
    SET_MASS_CALIBRATED = LINEAR_DETECTOR_VOLTAGE + 79
    SONAR_ENABLED = LINEAR_DETECTOR_VOLTAGE + 80
    QUAD_START_MASS = LINEAR_DETECTOR_VOLTAGE + 81
    QUAD_STOP_MASS = LINEAR_DETECTOR_VOLTAGE + 82
    QUAD_PEAK_WIDTH = LINEAR_DETECTOR_VOLTAGE + 83
    REFERENCE_SCAN = LINEAR_DETECTOR_VOLTAGE + 99
    USE_LOCKMASS_CORRECTION = LINEAR_DETECTOR_VOLTAGE + 100
    LOCKMASS_CORRECTION = LINEAR_DETECTOR_VOLTAGE + 101
    USETEMP_CORRECTION = LINEAR_DETECTOR_VOLTAGE + 102
    TEMP_CORRECTION = LINEAR_DETECTOR_VOLTAGE + 103
    TEMP_COEFFICIENT = LINEAR_DETECTOR_VOLTAGE + 104
    FAIMS_COMPENSATION_VOLTAGE = LINEAR_DETECTOR_VOLTAGE + 105
    TIC_TRACE_A = LINEAR_DETECTOR_VOLTAGE + 106
    TIC_TRACE_B = LINEAR_DETECTOR_VOLTAGE + 107
    RAW_EE_CV = LINEAR_DETECTOR_VOLTAGE + 108
    RAW_EE_CE = LINEAR_DETECTOR_VOLTAGE + 110
    ACCURATE_MASS = LINEAR_DETECTOR_VOLTAGE + 111
    ACCURATE_MASS_FLAGS = LINEAR_DETECTOR_VOLTAGE + 112
    SCAN_ERROR_FLAG = LINEAR_DETECTOR_VOLTAGE + 113
    DRE_TRANSMISSION = LINEAR_DETECTOR_VOLTAGE + 114
    SCAN_PUSH_COUNT = LINEAR_DETECTOR_VOLTAGE + 115
    RAW_STAT_SWAVE_NORMALISATION_FACTOR = LINEAR_DETECTOR_VOLTAGE + 116
    MIN_DRIFT_TIME_CHANNEL = LINEAR_DETECTOR_VOLTAGE + 121
    MAX_DRIFT_TIME_CHANNEL = LINEAR_DETECTOR_VOLTAGE + 122
    TOTAL_ION_CURRENT = LINEAR_DETECTOR_VOLTAGE + 251
    BASE_PEAK_MASS = LINEAR_DETECTOR_VOLTAGE + 252
    BASE_PEAK_INTENSITY = LINEAR_DETECTOR_VOLTAGE + 253
    PEAKS_IN_SCAN = LINEAR_DETECTOR_VOLTAGE + 254
    UNINITIALISED = LINEAR_DETECTOR_VOLTAGE + 298

class MassLynxSampleListItem(IntEnum):
    FILE_NAME = 700
    FILE_TEXT = FILE_NAME + 1
    MS_FILE = FILE_NAME + 2
    MS_TUNE_FILE = FILE_NAME + 3
    INLET_FILE = FILE_NAME + 4
    INLET_PRERUN = FILE_NAME + 5
    INLET_POSTRUN = FILE_NAME + 6
    INLET_SWITCH = FILE_NAME + 7
    AUTO_FILE = FILE_NAME + 8
    PROCESS = FILE_NAME + 9
    PROCESS_PARAMS = FILE_NAME + 10
    PROCESS_OPTIONS = FILE_NAME + 11
    ACQU_PROCESS_FILE = FILE_NAME + 12
    ACQU_PROCESS_PARAMS = FILE_NAME + 13
    ACQU_PROCESS_OPTIONS = FILE_NAME + 14
    PROCESS_ACTION = FILE_NAME + 15
    SAMPLE_LOCATION = FILE_NAME + 16
    SAMPLE_GROUP = FILE_NAME + 17
    JOB = FILE_NAME + 18
    TASK = FILE_NAME + 19
    USER = FILE_NAME + 20
    SUBMITTER = FILE_NAME + 21
    CONDITIONS = FILE_NAME + 22
    TYPE = FILE_NAME + 23
    CONTROL = FILE_NAME + 24
    ID = FILE_NAME + 25
    CONC_A = FILE_NAME + 26
    CONC_B = FILE_NAME + 27
    CONC_C = FILE_NAME + 28
    CONC_D = FILE_NAME + 29
    CONC_E = FILE_NAME + 30
    CONC_F = FILE_NAME + 31
    CONC_G = FILE_NAME + 32
    CONC_H = FILE_NAME + 33
    CONC_I = FILE_NAME + 34
    CONC_J = FILE_NAME + 35
    CONC_K = FILE_NAME + 36
    CONC_L = FILE_NAME + 37
    CONC_M = FILE_NAME + 38
    CONC_N = FILE_NAME + 39
    CONC_O = FILE_NAME + 40
    CONC_P = FILE_NAME + 41
    CONC_Q = FILE_NAME + 42
    CONC_R = FILE_NAME + 43
    CONC_S = FILE_NAME + 44
    CONC_T = FILE_NAME + 45
    WAVELENGTH_A = FILE_NAME + 46
    WAVELENGTH_B = FILE_NAME + 47
    WAVELENGTH_C = FILE_NAME + 48
    WAVELENGTH_D = FILE_NAME + 49
    WAVELENGTH_E = FILE_NAME + 50
    WAVELENGTH_F = FILE_NAME + 51
    WAVELENGTH_G = FILE_NAME + 52
    WAVELENGTH_H = FILE_NAME + 53
    WAVELENGTH_I = FILE_NAME + 54
    WAVELENGTH_J = FILE_NAME + 55
    MASS_A = FILE_NAME + 56
    MASS_B = FILE_NAME + 57
    MASS_C = FILE_NAME + 58
    MASS_D = FILE_NAME + 59
    MASS_E = FILE_NAME + 60
    MASS_F = FILE_NAME + 61
    MASS_G = FILE_NAME + 62
    MASS_H = FILE_NAME + 63
    MASS_I = FILE_NAME + 64
    MASS_J = FILE_NAME + 65
    MASS_K = FILE_NAME + 66
    MASS_L = FILE_NAME + 67
    MASS_M = FILE_NAME + 68
    MASS_N = FILE_NAME + 59
    MASS_O = FILE_NAME + 70
    MASS_P = FILE_NAME + 71
    MASS_Q = FILE_NAME + 72
    MASS_R = FILE_NAME + 73
    MASS_S = FILE_NAME + 74
    MASS_T = FILE_NAME + 75
    MASS_U = FILE_NAME + 76
    MASS_V = FILE_NAME + 77
    MASS_W = FILE_NAME + 78
    MASS_X = FILE_NAME + 79
    MASS_Y = FILE_NAME + 80
    MASS_Z = FILE_NAME + 81
    MASS_AA = FILE_NAME + 82
    MASS_BB = FILE_NAME + 83
    MASS_CC = FILE_NAME + 84
    MASS_DD = FILE_NAME + 85
    FRACTION_FILE = FILE_NAME + 86
    FRACTION_1 = FILE_NAME + 87
    FRACTION_2 = FILE_NAME + 88
    FRACTION_3 = FILE_NAME + 89
    FRACTION_4 = FILE_NAME + 90
    FRACTION_5 = FILE_NAME + 91
    FRACTION_6 = FILE_NAME + 92
    FRACTION_7 = FILE_NAME + 93
    FRACTION_8 = FILE_NAME + 94
    FRACTION_9 = FILE_NAME + 95
    FRACTION_10 = FILE_NAME + 96
    FRACTION_BOOLEAN_LOGIC = FILE_NAME + 97
    FRACTION_START = FILE_NAME + 98
    INJ_VOL = FILE_NAME + 99
    STOCK_DIL = FILE_NAME + 100
    USER_DIVISOR_1 = FILE_NAME + 101
    USER_FACTOR_1 = FILE_NAME + 102
    USER_FACTOR_2 = FILE_NAME + 103
    USER_FACTOR_3 = FILE_NAME + 104
    SPARE_1 = FILE_NAME + 105
    SPARE_2 = FILE_NAME + 106
    SPARE_3 = FILE_NAME + 107
    SPARE_4 = FILE_NAME + 108
    SPARE_5 = FILE_NAME + 109
    HPLC_FILE = FILE_NAME + 110
    QUAN_REF = FILE_NAME + 111
    AUTO_ADDITION = FILE_NAME + 112
    MOLFILE = FILE_NAME + 113
    SUBJECTTEXT = FILE_NAME + 114
    SUBJECTTIME = FILE_NAME + 115
    METH_DB = FILE_NAME + 116
    CURVE_DB = FILE_NAME + 117


class MassLynxScanType(IntEnum):
    MS1 = 1
    MS2 = MS1 + 2
    UNINITIALISED = MS1 + 9

class MassLynxAcquisitionType(IntEnum):
    DDA = 0
    MSE = DDA + 1
    HDDDA = DDA + 2
    HDMSE = DDA + 3
    SONAR = DDA + 4
    UNINITIALISED = DDA + 99

class ThresholdParameters(IntEnum):
    VALUE = 0
    THRESHTYPE = 1



class ThresholdType2(IntEnum):
    ABSOLUTE_1 = 0
    RELATIVE_2 = 1



class SmoothParameter(IntEnum):
    NUMBER = 1
    WIDTH = NUMBER + 1
    SMOOTHTYPE = NUMBER + 2


class SmoothType(IntEnum):
    MEAN = 1
    MEDIAN = MEAN + 1
    SAVITZKY_GOLAY = MEAN + 2



class LockMassParameter(IntEnum):
    MASS = 1000
    TOLERANCE = MASS + 1
    FORCE = MASS + 2


class FunctionDefinition(IntEnum):
    CONTINUUM = 1100
    IONMODE = CONTINUUM + 1
    FUNCTIONTYPE = CONTINUUM + 2
    STARTMASS = CONTINUUM + 3
    ENDMASS = CONTINUUM + 4
    CDT_SCANS = CONTINUUM + 5,
    SAMPLINGFREQUENCY = CONTINUUM + 6
    LTEFF = CONTINUUM + 7
    VEFF = CONTINUUM + 8,


class AnalogParameter(IntEnum):
    DESCRIPTION = 1200 + 1
    UNITS = DESCRIPTION + 2
    TYPE = DESCRIPTION + 3


class AnalogTraceType(IntEnum):
    ANALOG = 1250
    ELSD = ANALOG + 1
    READBACK = ANALOG + 2


class AutoLynxStatus(IntEnum):
    QUEUED = 1300
    PROCESSED = QUEUED + 1
    FAILED = QUEUED + 2
    NOTFOUND = QUEUED + 3
    UNINITIALISED = QUEUED + 9




import operator

from cpython cimport PyObject
from cpython.list cimport PyList_GetItem, PyList_Size, PyList_GET_SIZE, PyList_GET_ITEM, PyList_Append
from cpython.exc cimport PyErr_Occurred, PyErr_Clear, PyErr_ExceptionMatches
from cpython.dict cimport PyDict_GetItem, PyDict_SetItem, PyDict_Contains, PyDict_DelItem, PyDict_Next, PyDict_Values
from cpython.tuple cimport PyTuple_GetItem, PyTuple_Size
from cpython.set cimport PySet_Size

from ms_deisotope._c.peak_dependency_network.intervals cimport SpanningMixin, IntervalTreeNode

from ms_deisotope._c.feature_map.lcms_feature cimport FeatureBase, EmptyFeature
from ms_deisotope._c.feature_map.feature_fit cimport LCMSFeatureSetFit, map_coord

from ms_deisotope.peak_dependency_network.subgraph import GreedySubgraphSelection


cdef class SpanningMixin2D(object):
    """Provides methods for checking whether an entity
    which has a defined start and end point over a single
    dimension contains or overlaps with another entity in
    that same dimension.
    """

    def __init__(self, start, end):
        self.start = start
        self.end = end

    def __contains__(self, i):
        """Tests for point inclusion, `start <= i <= end`

        Parameters
        ----------
        i : Number
            The point to be tested

        Returns
        -------
        bool
        """
        return self._contains(i)

    cdef bint _contains(self, map_coord i):
        """Tests for point inclusion, `start <= i <= end`

        Parameters
        ----------
        i : map_coord
            The point to be tested

        Returns
        -------
        bool
        """
        return self.start.ge(i) and i.le(self.end)

    cpdef bint contains(self, map_coord i):
        """Tests for point inclusion, `start <= i <= end`

        Parameters
        ----------
        i : map_coord
            The point to be tested

        Returns
        -------
        bool
        """
        return self._contains(i)

    cpdef bint overlaps(self, SpanningMixin2D interval):
        """Tests whether another spanning entity with
        a defined start and end point overlaps with
        this spanning entity

        Parameters
        ----------
        interval : SpanningMixin

        Returns
        -------
        bool
        """
        cdef bint cond
        cond = (
            (self.start.le(interval.start) and self.end.ge(interval.start)) or \
            (self.start.ge(interval.start) and self.end.le(interval.end)) or \
            (self.start.ge(interval.start) and self.end.ge(interval.end) and self.start.ge(interval.end)) or \
            (self.start.le(interval.start) and self.end.ge(interval.start)) or \
            (self.start.le(interval.end) and self.end.ge(interval.end)))
        return cond

    cpdef bint is_contained_in_interval(self, SpanningMixin2D interval):
        """Tests whether this spanning entity is
        completely contained inside the other entity

        Parameters
        ----------
        interval : SpanningMixin

        Returns
        -------
        bool
        """
        return self.start.ge(interval.start) and self.end.le(interval.end)

    cpdef bint contains_interval(self, SpanningMixin2D interval):
        """Tests whether the other spanning entity is
        completely contained inside this entity

        Parameters
        ----------
        interval : SpanningMixin

        Returns
        -------
        bool
        """
        return self.start.le(interval.start) and self.end.ge(interval.end)



cpdef double score_getter(LCMSFeatureSetFit x):
    return x.score


cdef class FeatureNode(object):
    """Holds all the information about a single `LCMSFeature` instance
    and all the `LCMSFeatureSetFit` instances that depend upon it.

    Attributes
    ----------
    feature : LCMSFeature
    links : dict
        Mapping from feature fit to score
    """
    def __init__(self, feature, links=None):
        if links is None:
            links = {}
        self.feature = feature
        self.links = links
        self._hash = hash(self.feature)

    def __hash__(self):
        return self._hash

    def __reduce__(self):
        return self.__class__, (self.feature, self.links)

    def __eq__(self, FeatureNode other):
        if other is None:
            return False
        try:
            return self.feature == other.feature
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self == other

    def __contains__(self, fit):
        return fit in self.links

    def __repr__(self):
        return "%s(%s, %s)" % (self.__class__.__name__, self.feature, self.links)


cdef class DependenceCluster(SpanningMixin2D):
    def __init__(self, parent=None, dependencies=None, maximize=True):
        if parent is None:
            parent = self
        if dependencies is None:
            dependencies = []
        else:
            dependencies = sorted(dependencies, key=score_getter, reverse=maximize)
        self.parent = parent
        self.dependencies = dependencies
        self.best_fit = None
        self.maximize = maximize
        self._reset()

    cpdef _reset(self):
        self.start = self._start()
        self.end = self._end()
        self.best_fit = self._best_fit()

    cpdef add(self, LCMSFeatureSetFit fit):
        """
        Adds a new LCMSFeatureSetFit to this cluster, and ensures the sorted
        property still holds

        Parameters
        ----------
        fit : LCMSFeatureSetFit
            The fit being added to the record

        """
        self.dependencies.append(fit)
        self.dependencies.sort(key=score_getter, reverse=self.maximize)
        self._reset()

    cpdef disjoint_subset(self, int max_size=1000):
        cdef:
            list dependencies
        if len(self.dependencies) > max_size:
            dependencies = self.dependencies[:max_size]
        else:
            dependencies = self.dependencies

        graph = ConnectedSubgraph(dependencies, maximize=self.maximize)
        return graph.find_heaviest_path()

    cpdef list disjoint_best_fits(self, int max_size=1000):
        """
        Compute the best set of disjoint isotopic fits spanning this cluster

        Returns
        -------
        list of LCMSFeatureSetFit
        """
        fit_sets = tuple(self.disjoint_subset(max_size=max_size))
        best_fits = fit_sets
        return [node.fit for node in best_fits]

    cdef LCMSFeatureSetFit _best_fit(self):
        """
        Retrieve the absolute best single fit for the cluster

        Returns
        -------
        name : LCMSFeatureSetFit
            The best scoring LCMSFeatureSetFit in this cluster
        """
        return self.dependencies[0]

    cdef map_coord _start(self):
        """
        Determine the first mz-time coordinate for members of this cluster

        Returns
        -------
        map_coord
        """
        cdef:
            size_t i, n
            map_coord best, cur

        n = PyList_Size(self.dependencies)
        if n == 0:
            return None
        best = (<LCMSFeatureSetFit>PyList_GetItem(self.dependencies, 0)).get_start()
        for i in range(1, n):
            cur = (<LCMSFeatureSetFit>PyList_GetItem(self.dependencies, i)).get_start()
            if cur < best:
                best = cur
        return best.copy()

    cdef map_coord _end(self):
        """
        Determines the last mz-time coordinate for members of this cluster

        Returns
        -------
        map_coord
        """
        cdef:
            size_t i, n
            map_coord best, cur

        n = PyList_Size(self.dependencies)
        if n == 0:
            return None
        best = (<LCMSFeatureSetFit>PyList_GetItem(self.dependencies, 0)).get_end()
        for i in range(1, n):
            cur = (<LCMSFeatureSetFit>PyList_GetItem(self.dependencies, i)).get_end()

            if cur > best:
                best = cur
        return best.copy()

    def interval_contains_point(self, point):
        return self.start <= point < self.end

    def __len__(self):
        return len(self.dependencies)

    def __iter__(self):
        return iter(self.dependencies)

    def __contains__(self, fit):
        return fit in self.dependencies

    def __eq__(self, other):
        return self.dependencies == other.dependencies

    def __repr__(self):
        return "%s(dependencies=[%s], start=%r, end=%r)" % (
            self.__class__.__name__,
            '\n'.join(map(str, self.dependencies[:10])), self.start, self.end)

    def __getitem__(self, i):
        return self.dependencies[i]


cdef bint is_valid(FeatureBase feature):
    return feature is not None and not isinstance(feature, EmptyFeature)


cdef double INF = float('infinity')


cdef class FeatureSetFitNode(SpanningMixin2D):
    def __init__(self, fit, index=None):
        self.fit = fit
        self.edges = set()
        self.overlap_edges = set()
        self.index = -1
        if index is not None:
            self.index = index
        self._hash = -1
        self.score = fit.score
        self._init_indices()

    cdef void _init_indices(self):
        cdef:
            map_coord start, end
            set feature_indices
            size_t i, n
            double st, et
            FeatureBase feat

        feature_indices = set()
        start = map_coord._create(self.fit.get_start().mz, 0)
        end = map_coord._create(self.fit.get_end().mz, INF)

        # Build a minimal bounding box over features, one that allows overflow.
        n = PyList_Size(self.fit.features)
        for i in range(n):
            feat = <FeatureBase>PyList_GetItem(self.fit.features, i)
            if not is_valid(feat):
                continue
            st = feat.get_start_time()
            et = feat.get_end_time()
            feature_indices.add((feat.get_mz(), st, et))
            if start.time < st:
                start.time = st
            if end.time > et:
                end.time = et
        self.feature_indices = feature_indices
        self.start = start
        self.end = end

    def __hash__(self):
        if self._hash == -1:
            self._hash = hash(self.fit)
        return self._hash

    def __eq__(self, FeatureSetFitNode other):
        return self.fit is other.fit

    def __gt__(self, FeatureSetFitNode other):
        return self.fit > other.fit

    def __lt__(self, FeatureSetFitNode other):
        return self.fit < other.fit

    def __ne__(self, FeatureSetFitNode other):
        return self.fit is not other.fit

    cpdef visit(self, FeatureSetFitNode other):
        if self.isdisjoint(other):
            self.edges.add(other)
            other.edges.add(self)
        else:
            self.overlap_edges.add(other)
            other.overlap_edges.add(self)

    cpdef bint isdisjoint(self, FeatureSetFitNode other):
        return PySet_Size(self.feature_indices & other.feature_indices)

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.fit)

    @staticmethod
    def overlap(FeatureSetFitNode a, FeatureSetFitNode b):
        return (a.feature_indices & b.feature_indices)


cdef class ConnectedSubgraph(object):
    def __init__(self, fits, maximize=True):
        self.nodes = tuple(map(FeatureSetFitNode, fits))
        self.maximize = maximize
        self._init()

    cdef void _init(self):
        cdef:
            int i, n
            FeatureSetFitNode node

        n = PyTuple_Size(self.nodes)
        for i in range(n):
            node = <FeatureSetFitNode>PyTuple_GetItem(self.nodes, i)
            node.index = i
        self.populate_edges()

    def __getitem__(self, i):
        return self.nodes[i]

    def __iter__(self):
        return iter(self.nodes)

    def __len__(self):
        return len(self.nodes)

    cpdef populate_edges(self):
        cdef:
            size_t i, j, n
            FeatureSetFitNode node, other
        n = PyTuple_Size(self.nodes)
        for i in range(n):
            node = <FeatureSetFitNode>PyTuple_GetItem(self.nodes, i)
            for j in range(i + 1, n):
                other = <FeatureSetFitNode>PyTuple_GetItem(self.nodes, j)
                node.visit(other)

    cpdef find_heaviest_path(self, method="greedy"):
        if len(self) == 1:
            return set(self.nodes)
        if method == "greedy":
            solution = GreedySubgraphSelection.solve(
                self.nodes, maximize=self.maximize, overlap_fn=FeatureSetFitNode.overlap)
            return solution
        else:
            raise NotImplementedError(method)


cdef class FeatureDependenceGraphBase(object):
    cpdef list nodes_for(self, LCMSFeatureSetFit fit_record, dict cache=None):
        cdef:
            size_t i, n
            FeatureBase p
            list result
            PyObject* ptmp

        if cache is None:
            result = []
            n = PyList_Size(fit_record.features)
            for i in range(n):
                p = <FeatureBase>PyList_GET_ITEM(fit_record.features, i)
                if is_valid(p):
                    result.append(<object>PyDict_GetItem(self.nodes, p))

            return result
        else:
            ptmp = PyDict_GetItem(cache, fit_record)
            if ptmp != NULL:
                return <list>ptmp
            result = []
            n = PyList_Size(fit_record.features)
            for i in range(n):
                p = <FeatureBase>PyList_GET_ITEM(fit_record.features, i)
                if is_valid(p):
                    result.append(<object>PyDict_GetItem(self.nodes, p))
            PyDict_SetItem(cache, fit_record, result)
            return result

    cpdef add_fit_dependence(self, LCMSFeatureSetFit fit_record):
        cdef:
            size_t i, n
            FeatureBase p
            FeatureNode node
            PyObject* ptmp

        n = PyList_Size(fit_record.features)
        for i in range(n):
            p = <FeatureBase>PyList_GET_ITEM(fit_record.features, i)
            if not is_valid(p):
                continue
            node = <FeatureNode>PyDict_GetItem(self.nodes, p)
            PyDict_SetItem(node.links, fit_record, fit_record.score)
        self.dependencies.add(fit_record)

    cpdef drop_fit_dependence(self, LCMSFeatureSetFit fit_record, dict cache=None):
        cdef:
            list nodes
            size_t i, n
            int code
            FeatureNode node
        nodes = self.nodes_for(fit_record, cache=cache)
        n = PyList_Size(nodes)
        for i in range(n):
            node = <FeatureNode>PyList_GET_ITEM(nodes, i)
            code = PyDict_DelItem(node.links, fit_record)
            if code < 0:
                if PyErr_Occurred():
                    if PyErr_ExceptionMatches(KeyError):
                        PyErr_Clear()

    cpdef drop_superceded_fits(self):
        cdef:
            dict cache
            set suppressed, keep
            LCMSFeatureSetFit dep, candidate
            bint suppress, maximize
            FeatureBase monoisotopic_feature
            FeatureNode monoisotopic_feature_node
            object tmp
            PyObject* ptmp

        suppressed = set()
        keep = set()

        maximize = self.maximize
        for tmp in self.dependencies:
            dep = <LCMSFeatureSetFit>tmp
            monoisotopic_feature = dep.monoisotopic_feature
            if not is_valid(monoisotopic_feature):
                continue
            ptmp = PyDict_GetItem(self.nodes, monoisotopic_feature)
            if ptmp == NULL:
                raise KeyError(monoisotopic_feature)
            else:
                monoisotopic_feature_node = <FeatureNode>ptmp
            suppress = False
            for tmp in monoisotopic_feature_node.links:
                candidate = <LCMSFeatureSetFit>tmp
                if dep.charge == candidate.charge:
                    if monoisotopic_feature in candidate.features:
                        if (maximize and (dep.score < candidate.score)) or (not maximize and (dep.score > candidate.score)):
                            suppress = True
                            break
            if suppress:
                suppressed.add(dep)
            else:
                keep.add(dep)
        cache = dict()
        for dropped in suppressed:
            self.drop_fit_dependence(dropped, cache)
        self.dependencies = keep

    cpdef drop_gapped_fits(self, n=None):
        cdef:
            int max_missed
            LCMSFeatureSetFit dep
            set keep
        if n is None:
            max_missed = self.max_missing_features
        else:
            max_missed = n
        keep = set()
        for dep in self.dependencies:
            if dep.missing_features > max_missed:
                self.drop_fit_dependence(dep)
            else:
                keep.add(dep)
        self.dependencies = keep

    cpdef best_exact_fits(self):
        cdef:
            dict by_feature_set
            list bucket, fits
            set best_fits
            LCMSFeatureSetFit fit, best_fit
            tuple key
            PyObject* ptmp
            PyObject* pval
            Py_ssize_t pos
            size_t i, n
            bint maximize

        by_feature_set = {}
        best_fits = set()
        for tmp in self.dependencies:
            fit = <LCMSFeatureSetFit>tmp
            key = tuple(fit.features)
            ptmp = PyDict_GetItem(by_feature_set, key)
            if ptmp == NULL:
                bucket = []
                PyDict_SetItem(by_feature_set, key, bucket)
            else:
                bucket = <list>ptmp
            bucket.append(fit)

        maximize = self.maximize
        pos = 0
        while PyDict_Next(by_feature_set, &pos, &ptmp, &pval):
            fits = <list>pval
            n = PyList_Size(fits)
            if n == 1:
                best_fit = <LCMSFeatureSetFit>PyList_GET_ITEM(fits, 0)
                best_fits.add(best_fit)
                continue
            best_fit = <LCMSFeatureSetFit>PyList_GET_ITEM(fits, 0)
            for i in range(1, n):
                fit = <LCMSFeatureSetFit>PyList_GET_ITEM(fits, i)
                if (maximize and (best_fit.score < fit.score)) or (not maximize and (best_fit.score > fit.score)):
                    self.drop_fit_dependence(best_fit)
                    best_fit = fit
                else:
                    self.drop_fit_dependence(fit)
            best_fits.add(best_fit)
        self.dependencies = best_fits

    cpdef _gather_independent_clusters(self, dict nodes_for_cache=None):
        cdef:
            dict clusters
            set dependencies
            list dependencies_snapshot, dependent_nodes
            FeatureNode seed_node, dep_node
            list nodes, nodes_of
            LCMSFeatureSetFit dep
            Py_ssize_t i, n, j, m, k, p
            PyObject* ptemp

        clusters = dict()
        if nodes_for_cache is None:
            nodes_for_cache = dict()

        nodes = PyDict_Values(self.nodes)
        n = PyList_GET_SIZE(nodes)
        for i in range(n):
            seed_node = <FeatureNode>PyList_GET_ITEM(nodes, i)

            # This feature is depended upon by each fit in `dependencies`
            dependencies = set(seed_node.links)

            m = PySet_Size(dependencies)
            if m == 0:
                continue
            # These fits also depend upon these other peaks, and those peaks are depended upon
            # for other fits in turn, which depend upon the assignment of this peak.
            dependencies_snapshot = list(dependencies)
            dependent_nodes = []
            for j in range(m):
                dep = <LCMSFeatureSetFit>PyList_GET_ITEM(dependencies_snapshot, j)
                nodes_of = self.nodes_for(dep, nodes_for_cache)
                p = PyList_GET_SIZE(nodes_of)
                for k in range(p):
                    dep_node = <FeatureNode>PyList_GET_ITEM(nodes_of, k)
                    ptemp = PyDict_GetItem(clusters, dep_node)
                    if ptemp != NULL:
                        dependencies |= (<set>ptemp)
                    # Update all co-depended nodes with the full set of all fits which depend upon them
                    # PyDict_SetItem(clusters, dep_node, dependencies)
            for _dep in dependencies:
                dep = <LCMSFeatureSetFit>_dep
                nodes_of = self.nodes_for(dep, nodes_for_cache)
                p = PyList_GET_SIZE(nodes_of)
                for k in range(p):
                    dep_node = <FeatureNode>PyList_GET_ITEM(nodes_of, k)
                    PyDict_SetItem(clusters, dep_node, dependencies)

        return clusters



import warnings
import operator
from collections import defaultdict

from ms_deisotope.peak_dependency_network.intervals import (
    # Use pure Python implementations to allow non-primitive floats as coordinates
    _SpanningMixin as SpanningMixin, _IntervalTreeNode as IntervalTreeNode)
from ms_deisotope.peak_dependency_network.subgraph import GreedySubgraphSelection

from .lcms_feature import EmptyFeature
from .feature_fit import map_coord


def is_valid(feature):
    return feature is not None and not isinstance(feature, EmptyFeature)


class FeatureNode(object):
    """Holds all the information about a single `LCMSFeature` instance
    and all the `LCMSFeatureFit` instances that depend upon it.

    Attributes
    ----------
    feature : LCMSFeature
    links : dict
        Mapping from feature fit to score
    """
    def __init__(self, feature, links=None):
        if links is None:
            links = {}
        self.feature = feature
        self.links = links
        self._hash = hash(self.feature)

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        try:
            return self.feature == other.feature
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self == other

    def __contains__(self, fit):
        return fit in self.links

    def __repr__(self):
        return "FeatureNode(%s, %s)" % (self.feature, self.links)

try:
    _FeatureNode = FeatureNode
    has_c = True
    from ms_deisotope._c.feature_map.dependence_network import FeatureNode
except ImportError:
    has_c = False

class DependenceCluster(SpanningMixin):
    """
    Represent a cluster of isotopic fits which are overlapping

    Attributes
    ----------
    dependencies : set
        The set of all fits which are codependent
    start: float
        The smallest mz of all elements of this cluster
    end: float
        The largest mz of all elements of this cluster
    """

    def __init__(self, parent=None, dependencies=None, maximize=True):
        if parent is None:
            parent = self
        if dependencies is None:
            dependencies = []
        else:
            dependencies = sorted(dependencies, key=lambda x: x.score, reverse=maximize)
        self.parent = parent
        self.dependencies = dependencies
        self.best_fit = None
        self.maximize = maximize
        self._reset()

    def _reset(self):
        self.start = self._start()
        self.end = self._end()
        self.best_fit = self._best_fit()

    def add(self, fit):
        """
        Adds a new LCMSFeatureSetFit to this cluster, and ensures the sorted
        property still holds

        Parameters
        ----------
        fit : LCMSFeatureSetFit
            The fit being added to the record

        """
        self.dependencies.append(fit)
        self.dependencies.sort(key=lambda x: x.score, reverse=self.maximize)
        self._reset()

    def disjoint_subset(self, max_size=1000):
        if len(self.dependencies) > max_size:
            dependencies = self.dependencies[:max_size]
        else:
            dependencies = self.dependencies
        graph = ConnectedSubgraph(dependencies, maximize=self.maximize)
        return graph.find_heaviest_path()

    def _best_fit(self):
        """
        Retrieve the absolute best single fit for the cluster

        Returns
        -------
        name : LCMSFeatureSetFit
            The best scoring LCMSFeatureSetFit in this cluster
        """
        return self.dependencies[0]

    def disjoint_best_fits(self, max_size=1000):
        """
        Compute the best set of disjoint isotopic fits spanning this cluster

        Returns
        -------
        list of LCMSFeatureSetFit
        """
        fit_sets = tuple(self.disjoint_subset(max_size=max_size))
        best_fits = fit_sets
        return [node.fit for node in best_fits]

    def _start(self):
        """
        Determine the first mz coordinate for members of this cluster

        Returns
        -------
        float
        """
        return map_coord(*min([member.start for member in self.dependencies]))

    def _end(self):
        """
        Determines the last mz coordinate for members of this cluster

        Returns
        -------
        float
        """
        return map_coord(*max([member.end for member in self.dependencies]))

    def interval_contains_point(self, point):
        return self.start <= point < self.end

    def __len__(self):
        return len(self.dependencies)

    def __iter__(self):
        return iter(self.dependencies)

    def __contains__(self, fit):
        return fit in self.dependencies

    def __eq__(self, other):
        return self.dependencies == other.dependencies

    def __repr__(self):
        return "DependenceCluster(dependencies=[%s], start=%r, end=%r)" % (
            '\n'.join(map(str, self.dependencies[:10])), self.start, self.end)

    def __getitem__(self, i):
        return self.dependencies[i]


try:
    _DependenceCluster = DependenceCluster
    has_c = True
    from ms_deisotope._c.feature_map.dependence_network import DependenceCluster
except ImportError:
    has_c = False


class FeatureSetFitNode(SpanningMixin):
    def __init__(self, fit, index=None):
        self.fit = fit
        self.edges = set()
        self.overlap_edges = set()

        self._hash = None
        self.score = fit.score

        self.feature_indices = {(f.mz, f.start_time, f.end_time) for f in fit.features if is_valid(f)}

        self.start = (fit.start[0], max(f.start_time for f in fit.features if is_valid(f)))
        self.end = (fit.end[0], min(f.end_time for f in fit.features if is_valid(f)))

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(self.fit)
        return self._hash

    def __eq__(self, other):
        return self.fit is other.fit

    def __gt__(self, other):
        return self.fit > other.fit

    def __lt__(self, other):
        return self.fit < other.fit

    def __ne__(self, other):
        return self.fit is not other.fit

    def visit(self, other):
        if self.isdisjoint(other):
            self.edges.add(other)
            other.edges.add(self)
        else:
            self.overlap_edges.add(other)
            other.overlap_edges.add(self)

    def isdisjoint(self, other):
        return self.feature_indices.isdisjoint(other.feature_indices)

    def __repr__(self):
        return "FeatureSetFitNode(%r)" % self.fit

    @staticmethod
    def overlap(a, b):
        return len(a.feature_indices & b.feature_indices) > 0


try:
    _FeatureSetFitNode = FeatureSetFitNode
    has_c = True
    from ms_deisotope._c.feature_map.dependence_network import FeatureSetFitNode
except ImportError:
    has_c = False


class ConnectedSubgraph(object):
    def __init__(self, fits, maximize=True):
        self.nodes = tuple(map(FeatureSetFitNode, fits))
        self.maximize = maximize
        i = 0
        for node in self.nodes:
            node.index = i
            i += 1
        self.populate_edges()

    def __getitem__(self, i):
        return self.nodes[i]

    def __iter__(self):
        return iter(self.nodes)

    def __len__(self):
        return len(self.nodes)

    def populate_edges(self):
        for i in range(len(self.nodes)):
            node = self.nodes[i]
            for j in range(i + 1, len(self.nodes)):
                other = self.nodes[j]
                node.visit(other)

    def find_heaviest_path(self, method="greedy"):
        if len(self) == 1:
            return set(self.nodes)
        if method == "greedy":
            solution = GreedySubgraphSelection.solve(
                tuple(self), maximize=self.maximize, overlap_fn=FeatureSetFitNode.overlap)
            return solution
        else:
            raise NotImplementedError(method)


try:
    _ConnectedSubgraph = ConnectedSubgraph
    has_c = True
    from ms_deisotope._c.feature_map.dependence_network import ConnectedSubgraph
except ImportError:
    has_c = False


class FeatureDependenceGraphBase(object):

    def nodes_for(self, fit_record, cache=None):
        if cache is None:
            return [self.nodes[p] for p in fit_record.features if is_valid(p)]
        else:
            try:
                return cache[fit_record]
            except KeyError:
                cache[fit_record] = value = [
                    self.nodes[p] for p in fit_record.features if is_valid(p)]
                return value

    def add_fit_dependence(self, fit_record):
        for feature in fit_record.features:
            if not is_valid(feature):
                continue
            self.nodes[feature].links[fit_record] = fit_record.score
        self.dependencies.add(fit_record)

    def drop_superceded_fits(self):
        suppressed = []
        keep = []

        if self.maximize:
            comparator = operator.lt
        else:
            comparator = operator.gt

        for dep in self.dependencies:
            monoisotopic_feature = dep.monoisotopic_feature
            if not is_valid(monoisotopic_feature):
                continue
            monoisotopic_feature_node = self.nodes[monoisotopic_feature]
            suppress = False
            for candidate in monoisotopic_feature_node.links:
                if dep.charge == candidate.charge:
                    if monoisotopic_feature in candidate.features:
                        if comparator(dep.score, candidate.score):
                            suppress = True
                            break
            if suppress:
                suppressed.append(dep)
            else:
                keep.append(dep)
        for dropped in suppressed:
            self.drop_fit_dependence(dropped)
        self.dependencies = set(keep)

    def drop_fit_dependence(self, fit_record, cache=None):
        for node in self.nodes_for(fit_record, cache=cache):
            try:
                del node.links[fit_record]
            except KeyError:
                pass

    def drop_gapped_fits(self, n=None):
        if n is None:
            n = self.max_missing_features
        keep = []
        for dep in self.dependencies:
            if dep.missing_features > n:
                self.drop_fit_dependence(dep)
            else:
                keep.append(dep)
        self.dependencies = set(keep)

    def best_exact_fits(self):
        by_feature_set = defaultdict(list)
        best_fits = []
        for fit in self.dependencies:
            by_feature_set[tuple(fit.features)].append(fit)
        for peak_tuple, fits in by_feature_set.items():
            fits = sorted(fits, key=lambda x: x.score,
                          reverse=not self.maximize)
            for fit in fits[:-1]:
                self.drop_fit_dependence(fit)
            best_fits.append(fits[-1])
        self.dependencies = set(best_fits)

    def _gather_independent_clusters(self, nodes_for_cache=None):
        clusters = defaultdict(set)
        if nodes_for_cache is None:
            nodes_for_cache = {}

        for node in self.nodes.values():
            # This feature is depended upon by each fit in `dependencies`
            dependencies = set(node.links.keys())

            if len(dependencies) == 0:
                continue

            # These fits also depend upon these other features, and those features are depended upon
            # for other fits in turn, which depend upon the assignment of this feature.
            for dep in list(dependencies):
                for node in self.nodes_for(dep, nodes_for_cache):
                    dependencies |= clusters[node]

            # Create a fresh copy to share out again to avoid eliminate possible errors of shared storage.
            # This is likely unecessary under the current implementation.
            dependencies = set(dependencies)

            # Update all co-depended nodes with the full set of all fits which depend upon them
            for dep in dependencies:
                for node in self.nodes_for(dep, nodes_for_cache):
                    clusters[node] = dependencies
        return clusters


try:
    _FeatureDependenceGraphBase = FeatureDependenceGraphBase
    has_c = True
    from ms_deisotope._c.feature_map.dependence_network import FeatureDependenceGraphBase
except ImportError:
    has_c = False


class FeatureDependenceGraph(FeatureDependenceGraphBase):
    def __init__(self, feature_map, nodes=None, dependencies=None, max_missing_features=1,
                 use_monoisotopic_superceded_filtering=True, maximize=True):
        if nodes is None:
            nodes = {}
        if dependencies is None:
            dependencies = set()
        self.feature_map = feature_map
        self.nodes = nodes
        self.dependencies = dependencies
        self.max_missing_features = max_missing_features
        self.use_monoisotopic_superceded_filtering = use_monoisotopic_superceded_filtering
        self.maximize = maximize
        if len(self.nodes) == 0:
            self._populate_initial_graph()
        self.clusters = None
        self._interval_tree = None
        self._solution_map = {}

    def reset(self):
        self.nodes = dict()
        self.dependencies = set()
        self._interval_tree = None
        self._solution_map = dict()
        self._populate_initial_graph()

    def add_solution(self, key, solution):
        self._solution_map[key] = solution

    @property
    def interval_tree(self):
        if self._interval_tree is None:
            self._interval_tree = IntervalTreeNode.build(self.clusters)
            if self._interval_tree is None:
                raise ValueError(
                    "Could not build intervals for peak retrieval with %d clusters" % len(
                        self.clusters), self)
        return self._interval_tree

    def _deep_fuzzy_solution_for(self, peak, shift=0.5):
        pass

    def _find_fuzzy_solution_for(self, feature, shift=0.5):
        tree = self.interval_tree
        clusters = tree.contains_point(feature.mz + shift)
        if len(clusters) == 0:
            return None
        else:
            best_fits = [cluster.disjoint_best_fits() for cluster in clusters]

            acc = []
            for fits in best_fits:
                acc.extend(fits)
            best_fits = acc

            index = 0
            error = float('inf')

            for i, fit in enumerate(best_fits):
                err = abs(fit.monoisotopic_feature.mz - feature.mz)
                if err < error:
                    error = err
                    index = i
            fit = best_fits[index]
            return self._solution_map[fit]

    def find_solution_for(self, feature):
        feature_node = self.nodes[feature]
        tree = self.interval_tree
        if tree is None:
            tree = IntervalTreeNode.build(self.clusters)
            if tree is None:
                raise ValueError(
                    "Could not build intervals for peak retrieval with %d clusters" % len(
                        self.clusters))
        clusters = tree.contains_point(feature.mz)
        if len(clusters) == 0:
            return self._find_fuzzy_solution_for(feature)
        best_fits = [cluster.disjoint_best_fits() for cluster in clusters]

        acc = []
        for fits in best_fits:
            acc.extend(fits)
        best_fits = acc

        common = tuple(set(best_fits) & set(feature_node.links))

        if len(common) > 1 or len(common) == 0:
            if len(common) > 1:
                warnings.warn("Too many solutions exist for %r" % feature)
            # If there were no fits for this peak, then it may be that this peak
            # was not included in a fit. Try to find the nearest solution.
            i = 0
            err = float('inf')
            for j, case in enumerate(best_fits):
                case_err = abs(case.monoisotopic_feature.mz - feature.mz)
                if case_err < err:
                    i = j
                    err = case_err
            fit = best_fits[i]
        else:
            fit = common[0]
        return self._solution_map[fit]

    def _populate_initial_graph(self):
        for feature in self.feature_map:
            self.nodes[feature] = FeatureNode(feature)

    def find_non_overlapping_intervals(self):
        self.drop_gapped_fits()
        self.best_exact_fits()
        if self.use_monoisotopic_superceded_filtering:
            self.drop_superceded_fits()

        nodes_for_cache = {}

        clusters = self._gather_independent_clusters(nodes_for_cache)

        # Use an `id` keyed dictionary over these copied sets to ensure we have exactly one reference
        # to each set of inter-dependent fits, and then convert each set into an instance of `DependenceCluster`
        clusters = [DependenceCluster(dependencies=c, maximize=self.maximize) for c in {
            id(v): v for v in clusters.values() if v}.values()]
        clusters = sorted(clusters, key=operator.attrgetter("start"))
        self.clusters = clusters
        return clusters

    def __iter__(self):
        return iter(self.clusters)

    def __getitem__(self, i):
        return self.clusters[i]

    def __repr__(self):
        return "%s(%s, %d)" % (self.__class__.__name__, self.feature_map, len(self.dependencies))



from __future__ import print_function
import operator
import random

from datetime import datetime
from collections import OrderedDict

from six import add_metaclass


try:  # pragma: no cover
    from .plot import (
        draw_peaklist, draw_raw, annotate_scan, annotate_scan_single,
        has_plot)
except (RuntimeError, ImportError):  # pragma: no cover
    has_plot = False

    def annotate_scan(scan, products, nperrow=4, ax=None):
        raise ImportError('matplotlib')

    def annotate_scan_single(*args, **kwargs):
        raise ImportError('matplotlib')

try:
    range = xrange
except NameError:
    range = range


try:  # pragma: no cover
    i128 = long
    basestring = basestring
except NameError:  # pragma: no cover
    i128 = int
    basestring = (bytes, str)


def printer(message):
    print(datetime.now().isoformat(' ') + ' ' + str(message))


def debug_printer(message):
    print("DEBUG:" + datetime.now().isoformat(' ') + ' ' + str(message))


def simple_repr(self):  # pragma: no cover
    """A convenient function for automatically generating a ``__repr__``-like
    string for arbitrary objects.

    Returns
    -------
    str
    """
    template = "{self.__class__.__name__}({d})"

    def formatvalue(v):
        if isinstance(v, float):
            return "%0.4f" % v
        else:
            return str(v)

    if not hasattr(self, "__slots__") or len(self.__slots__) == 0 or hasattr(self, '__dict__'):
        d = [
            "%s=%s" % (k, formatvalue(v)) if v is not self else "(...)" for k, v in sorted(
                self.__dict__.items(), key=lambda x: x[0])
            if (not k.startswith("_") and not callable(v)) and not (v is None)]
    else:
        d = [
            "%s=%s" % (k, formatvalue(v)) if v is not self else "(...)" for k, v in sorted(
                [(name, getattr(self, name)) for name in self.__slots__], key=lambda x: x[0])
            if (not k.startswith("_") and not callable(v)) and not (v is None)]

    return template.format(self=self, d=', '.join(d))


class Base(object):
    """A convenience base class for non-critical code to provide types
    with automatic :meth:`__repr__` methods using :func:`simple_repr`
    """
    __slots__ = ()
    __repr__ = simple_repr


class Constant(object):
    """A convenience type meant to be used to instantiate singletons for signaling
    specific states in return values.

    Attributes
    ----------
    name: str
        The name of the constant
    """
    def __init__(self, name, is_true=True):
        self.name = name
        self.is_true = is_true

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return self.name == str(other)

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return str(self.name)

    def __nonzero__(self):
        return self.is_true

    def __bool__(self):
        return self.is_true

    def __contains__(self, i):
        return i in self.name

    def __getitem__(self, i):
        return self.name[i]

    def __iter__(self):
        return iter(self.name)

    def __len__(self):
        return len(self.name)


class DeconvolutionProcessResult(object):
    """Hold information from a multi-stage deconvolution process.

    Emulates an interface matching a tuple of (:attr:`peak_set`, :attr:`priorities`)

    Attributes
    ----------
    deconvoluter : :class:`~.DeconvoluterBase`
        The deconvoluter used.
    errors : list, optional
        A list of :class:`~.Exception` instances that may have been
        thrown during any stage of deconvolution.
    peak_set : :class:`~.DeconvolutedPeakSet`
        The resulting deconvoluted peaks
    priorities : list
        The extracted results from the targeted deconvolution list
    """

    def __init__(self, deconvoluter, peak_set, priorities, errors=None):
        self.deconvoluter = deconvoluter
        self.peak_set = peak_set
        self.priorities = priorities
        self.errors = errors

    def __getitem__(self, i):
        if i == 0:
            return self.peak_set
        elif i == 1:
            return self.priorities
        else:
            raise IndexError(i)

    def __iter__(self):
        yield self.peak_set
        yield self.priorities

    def __repr__(self):
        return "DeconvolutionProcessResult(%s, %s)" % tuple(self)


class TargetedDeconvolutionResultBase(Base):
    """Base class to store all of the necessary information to retrieve the optimal
    solution for a single peak.

    Attributes
    ----------
    deconvoluter : :class:`~.DeconvoluterBase`
        The deconvoluter to use to look up the result
    """
    def __init__(self, deconvoluter, *args, **kwargs):
        self.deconvoluter = deconvoluter

    def get(self):
        """Fetch the optimal solution after the computation has finished.

        Returns
        -------
        :class:`~.DeconvolutedPeak`
        """
        raise NotImplementedError()


class TrivialTargetedDeconvolutionResult(TargetedDeconvolutionResultBase):
    """Stores the necessary information to retrieve the local optimal solution for
    a single peak for a deconvolution algorithm where the local optimum is the best
    solution.

    Attributes
    ----------
    query_peak : :class:`~.FittedPeak`
        The peak queried with
    solution_peak : :class:`~.DeconvolutedPeak`
        The optimal solution peak
    """
    def __init__(self, deconvoluter, solution_peak, query_peak, *args, **kwargs):
        super(TrivialTargetedDeconvolutionResult, self).__init__(deconvoluter, *args, **kwargs)
        self.solution_peak = solution_peak
        self.query_peak = query_peak

    def get(self):
        """Fetch the optimal solution after the computation has finished.

        Returns
        -------
        :class:`~.DeconvolutedPeak`
        """
        return self.solution_peak


def ppm_error(x, y):
    return (x - y) / y


def ppm2da(mass, error_tolerance):
    x = mass + mass * error_tolerance
    return x - mass


def da2ppm(mass, error_tolerance):
    x = mass + error_tolerance
    return (x - mass) / mass


def dict_proxy(attribute):
    """Return a decorator for a class to give it a `dict`-like API proxied
    from one of its attributes

    Parameters
    ----------
    attribute : str
        The string corresponding to the attribute which will
        be used

    Returns
    -------
    function
    """
    getter = operator.attrgetter(attribute)

    def wrap(cls):

        def __getitem__(self, key):
            return getter(self)[key]

        def __setitem__(self, key, value):
            d = getter(self)
            d[key] = value

        def keys(self):
            return getter(self).keys()

        def __iter__(self):
            return iter(getter(self))

        def items(self):
            return getter(self).items()

        cls.__getitem__ = __getitem__
        cls.__setitem__ = __setitem__
        cls.keys = keys
        cls.items = items
        cls.__iter__ = __iter__

        return cls
    return wrap


def uid(n=128):
    """Generate a random "universally unique" ID number with ``n``
    bits of entropy.

    Parameters
    ----------
    n: int
        The number of random bits to generate

    Returns
    -------
    int
    """
    int_ = random.getrandbits(n)
    return int_


class LRUDict(object):  # pragma: no cover
    def __init__(self, *args, **kwargs):
        maxsize = kwargs.pop("maxsize", 24)
        self.store = OrderedDict()
        self.maxsize = maxsize
        self.purge()

    def __len__(self):
        return len(self.store)

    def popitem(self, last=True):
        return self.store.popitem(last=last)

    def pop(self, key, default=None):
        return self.store.pop(key, default)

    def purge(self):
        overflow = max(0, len(self) - self.maxsize)
        for _ in range(overflow):
            self.popitem(last=False)

    def __repr__(self):
        return "LRUDict(%r)" % (dict(self.store),)

    def __contains__(self, key):
        return key in self.store

    def __iter__(self):
        return iter(self.store)

    def keys(self):
        return self.store.keys()

    def values(self):
        return self.store.values()

    def items(self):
        return self.store.items()

    def __getitem__(self, key):
        value = self.store[key]
        self._mark_used(key)
        return value

    def __setitem__(self, key, value):
        self.store[key] = value
        self.purge()

    def _mark_used(self, key):
        value = self.store.pop(key, None)
        self.store[key] = value


def decimal_shift(x):
    i = 1.0
    while i < 10 ** 10:
        if round(x * i) > 0:
            return 1.0 / i
        i *= 10.
    return 1.0 / i


class _MappingOverAttributeProxy(object):
    """A replacement for __dict__ for unpickling an object which once
    has __slots__ now but did not before."""

    def __init__(self, obj):
        self.obj = obj

    def __getitem__(self, key):
        return getattr(self.obj, key)

    def __setitem__(self, key, value):
        setattr(self.obj, key, value)

    def __contains__(self, key):
        return hasattr(self.obj, key)

    def __repr__(self):
        return "{self.__class__.__name__}({self.obj})".format(self=self)



import copy
import ctypes
from ctypes import*
from array import *
from .MassLynxRawReader import *

class MassLynxSampleList(object):
    """Class to create MassLynx Sample lists"""

    def __init__(self):
        self._mlSampleList = c_void_p()              # instance variable
        self._codeHandler = MassLynxCodeHandler()
        self._stringHandler = MassLynxStringHandler()
        createSampleList = MassLynxRawReader.massLynxDll.createSampleList
        createSampleList.argtypes = [POINTER(c_void_p)]
        self.CheckReturnCode(createSampleList(self._mlSampleList))

    # destroy the samplelist
    def __del__(self):
        destroySampleList = MassLynxRawReader.massLynxDll.destroySampleList
        destroySampleList.argtypes = [c_void_p]
        destroySampleList( self._mlSampleList )

    # handle the retun code - no exceptions
    def CheckReturnCode( self, code ):
        return self._codeHandler.CheckReturnCode(code, False)

    # get the error message
    def GetLastMessage(self):
        return self._codeHandler.GetLastMessage()

    # get the error code
    def GetLastCode(self):
        return self._codeHandler.GetLastCode()

    # get the sample list in csv
    def ToString( self ):
        temp = c_char_p()
        sampleListToString = MassLynxRawReader.massLynxDll.sampleListToString
        sampleListToString.argtypes = [c_void_p, POINTER(c_char_p)]
        self.CheckReturnCode( sampleListToString(self._mlSampleList, temp) )
        return self._stringHandler.ToString( temp, False )

    def AddRow( self, row ):
        addSampleListRow = MassLynxRawReader.massLynxDll.addSampleListRow
        addSampleListRow.argtypes = [c_void_p, c_void_p]
        return self.CheckReturnCode( addSampleListRow(self._mlSampleList, row.GetParameters()) )



cimport cython

from ms_peak_picker._c.peak_set cimport PeakSet, PeakSetIndexed, FittedPeak
from ms_peak_picker._c.peak_index cimport PeakIndex

from brainpy._c.isotopic_distribution cimport TheoreticalPeak

from ms_deisotope._c.scoring cimport IsotopicFitterBase, IsotopicFitRecord
from ms_deisotope._c.averagine cimport AveragineCache, TheoreticalIsotopicPattern

cimport numpy as np


ctypedef fused FittedPeakCollection:
    PeakSet
    PeakIndex


ctypedef fused FittedPeakOrPosition:
    FittedPeak
    size_t
    long


cdef class DeconvoluterBase(object):
    cdef:
        public bint use_subtraction
        public str scale_method
        public bint merge_isobaric_peaks
        public double minimum_intensity

        public PeakSet peaklist

        public IsotopicFitterBase scorer
        public bint verbose
        public dict _slice_cache

    cpdef PeakSet between(self, double m1, double m2)
    cpdef FittedPeak has_peak(self, double mz, double error_tolerance)
    cdef FittedPeak _has_peak(self, double mz, double error_tolerance)

    cpdef list match_theoretical_isotopic_distribution(self, list theoretical_distribution, double error_tolerance=*)
    cpdef scale_theoretical_distribution(self, TheoreticalIsotopicPattern theoretical_distribution, list experimental_distribution)
    cpdef IsotopicFitRecord _evaluate_theoretical_distribution(self, list experimental, TheoreticalIsotopicPattern theoretical, FittedPeak peak, int charge)
    cpdef subtraction(self, TheoreticalIsotopicPattern isotopic_cluster, double error_tolerance=*)
    cpdef list fit_incremental_truncation(self, IsotopicFitRecord seed_fit, double lower_bound)

    cpdef bint _check_fit(self, IsotopicFitRecord fit)

    cpdef list _find_next_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)
    cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)

    cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)
    cdef int _find_previous_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)


cdef class AveragineDeconvoluterBase(DeconvoluterBase):
    cdef:
        public AveragineCache averagine

    cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,
                                                         double charge_carrier=*, double truncate_after=*,
                                                         double ignore_below=*)

    cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=*, double truncate_after=*,
                                    double ignore_below=*)


cdef class MultiAveragineDeconvoluterBase(DeconvoluterBase):
    cdef:
        public list averagines

    cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,
                                                         AveragineCache  averagine, double charge_carrier=*, double truncate_after=*,
                                                         double ignore_below=*)
    cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=*, double truncate_after=*,
                                    double ignore_below=*)


cdef bint has_multiple_real_peaks(list peaklist)


cpdef set _get_all_peak_charge_pairs(DeconvoluterBase self, FittedPeak peak, double error_tolerance=*,
                                     object charge_range=*,
                                     int left_search_limit=*, int right_search_limit=*,
                                     bint recalculate_starting_peak=*, bint use_quick_charge=*)

cpdef np.ndarray[np.int32_t, ndim=1] quick_charge(FittedPeakCollection peak_set, FittedPeakOrPosition position_spec, int min_charge, int max_charge)


cdef class ChargeIterator(object):
    cdef:
        public int lower
        public int upper
        public int sign
        int* values
        public size_t size
        public size_t index


    @staticmethod
    cdef ChargeIterator _create(int lo, int hi)

    @staticmethod
    cdef ChargeIterator from_tuple(tuple charge_range)

    @staticmethod
    cdef ChargeIterator from_quickcharge(tuple charge_range, PeakSet peaks, FittedPeak peak)

    cpdef sequence_from_quickcharge(self, PeakSet peaks, FittedPeak peak)
    cpdef make_sequence(self)

    cdef void release_sequence(self)

    cdef void set_bounds(self, int lo, int hi)

    cpdef bint has_more(self)
    cpdef reset(self)
    cpdef int get_next_value(self)

    cdef size_t get_size(self)



import re
import functools
import io
import os

from typing import Optional, Union

import numpy as np

from ms_deisotope.data_source.scan.base import RawDataArrays, ChargeNotProvided, PrecursorInformation
from ms_deisotope.data_source.memory import make_scan
from ms_deisotope.data_source._compression import get_opener
from ms_deisotope.data_source.scan.scan import WrappedScan


def scan_from_csv(file_handle: Union[io.IOBase, os.PathLike], delimiter: str=',', ms_level: int=2,
                  is_profile: bool=True, polarity: int=1, precursor_mz: Optional[float]=None,
                  precursor_charge: Optional[int]=None, skiprows: Optional[int]=None) -> WrappedScan:
    """Read an m/z-intensity point list from a text file stream.

    Parameters
    ----------
    file_handle : file
        The file to read from.
    delimiter : str, optional
        The field separator between m/z and intensity. (the default is ','). The
        delimiter can be a regular expression or escape sequence.
    ms_level : int, optional
        The MS level of the read scan. (the default is 2)
    is_profile : bool, optional
        Whether the scan is in profile mode (the default is True)
    polarity : int, optional
        Whether the scan is positive mode or negative (the default is 1, which means positive)
    precursor_mz : float, optional
        The m/z of the precursor ion to record. If provided, it will be
        provided in :attr:`~.Scan.precursor_information`.
    precursor_charge : int, optional
        If provided, and `precursor_mz` is not :const:`None`, then it will be provided
        in :attr:`~.Scan.precursor_information`.
    skiprows : int, optional
        The number of rows to skip before starting to parse data, if there is a more elaborate
        header.

    Returns
    -------
    :class:`~.Scan`
    """
    if skiprows is None:
        skiprows = 0
    file_handle = get_opener(file_handle)
    mzs = []
    intensities = []
    tokenizer = re.compile(delimiter)
    for i, line in enumerate(file_handle):
        if i < skiprows:
            continue
        line = tokenizer.split(line.strip())
        if i == skiprows:
            try:
                float(line[0])
            except Exception:
                continue
        mzs.append(line[0])
        intensities.append(line[1])
    mzs = np.array(mzs, dtype=float)
    intensities = np.array(intensities, dtype=float)
    signal = RawDataArrays(mzs, intensities)
    pinfo = None
    if precursor_mz is not None:
        pinfo = PrecursorInformation(
            precursor_mz,
            0,
            ChargeNotProvided if precursor_charge is None else precursor_charge)
    scan = make_scan(
        signal, ms_level, "index=1", 0, 0, is_profile=is_profile,
        polarity=polarity,
        precursor_information=pinfo)
    return scan

scan_from_table = functools.partial(scan_from_csv, delimiter=r'\s+')



from collections import defaultdict, namedtuple

from .similarity_methods import SpectrumAlignment

Neighbor = namedtuple("Neighbor", ('score', 'shift', 'scan_id', 'weight'))


class SpectrumAlignmentGraph(object):
    """A graph of spectra that can find similar pairs of shifted spectra from
    the population, constructing a support network for each spectrum.

    Attributes
    ----------
    scans : list of :class:`~.ScanBase`
        The spectra to be related.
    threshold : float
        The minimum alignment similarity to accept for supporters
    error_tolerance : float
        The ppm error tolerance to use when matching peaks, defaults to 2e-5
    min_delta : float
        The minimum precursor mass delta recognized. Below this number and the
        precursor mass delta is assumed to be the same and is skipped.
    match_charge : bool
        Whether or not to require charge states to match to construct an edge.
    """

    def __init__(self, scans, threshold=0.5, error_tolerance=2e-5, min_delta=0.01, match_charge=True):
        self.scans = list(scans)
        self.threshold = threshold
        self.error_tolerance = error_tolerance
        self.min_delta = min_delta
        self.match_charge = match_charge

        self.edges = defaultdict(dict)
        self.supporters = dict()

        self.build_edges()
        self.trim()

    def build_edges(self):
        edges = defaultdict(dict)
        self.scans.sort(key=lambda x: x.precursor_information.neutral_mass)
        for i, scan1 in enumerate(self.scans):
            for scan2 in self.scans[i + 1:]:
                if scan1 is scan2:
                    continue
                if self.match_charge and scan1.precursor_information.charge != scan2.precursor_information.charge:
                    continue
                delta = scan2.precursor_information.neutral_mass - \
                    scan1.precursor_information.neutral_mass
                if abs(delta) < self.min_delta:
                    continue
                aln = SpectrumAlignment(
                    scan1.peaks(), scan2.peaks(),
                    error_tolerance=self.error_tolerance,
                    shift=scan2.precursor_information.neutral_mass - scan1.precursor_information.neutral_mass)
                weight = sum([pp.score for pp in aln.peak_pairs])
                edges[scan1.id][scan2.id] = (aln.score, aln.shift, weight)
                edges[scan2.id][scan1.id] = (aln.score, -aln.shift, weight)
        self.edges = edges

    def trim(self):
        supporters = {}
        for scan1_id, neighbors in self.edges.items():
            buckets = {}
            for scan2_id, (score, shift, weight) in neighbors.items():
                if score < self.threshold:
                    continue
                key = int(shift * 100.0)
                if key in buckets and buckets[key][0] < score:
                    buckets[key] = Neighbor(score, shift, scan2_id, weight)
                else:
                    buckets[key] = Neighbor(score, shift, scan2_id, weight)
            supporters[scan1_id] = list(buckets.values())
        self.supporters = supporters

    def average_shifts(self):
        seen = set()

        shifts = []

        for a, dst in self.edges.items():
            for b, ed in dst.items():
                key = frozenset((a, b))
                if key in seen:
                    continue
                seen.add(key)
                (score, shift, weight) = ed
                if score < self.threshold:
                    continue
                shift = abs(shift)
                shifts.append((shift, weight, score))

        shifts.sort(key=lambda x: x[1], reverse=True)

        bins = []
        for shift in shifts:
            for sbin in bins:
                if sbin.test(shift[0], 0.1):
                    sbin.add(shift)
                    break
            else:
                bins.append(ShiftBin(shift[0], [shift]))

        bins.sort(key=lambda x: x.average())
        return bins

    def to_json(self):
        container = {
            "edges": dict(self.edges),
            "supporters": self.supporters,
            "scan_ids": [s.id for s in self.scans],
            "parameters": {
                "threshold": self.threshold,
                "error_tolerance": self.error_tolerance,
                "min_delta": self.min_delta,
                "match_charge": self.match_charge
            }
        }
        return container


class ShiftBin(object):
    def __init__(self, mass, observations=None):
        if observations is None:
            observations = []
        self.mass = mass
        self.observations = observations

    def add(self, observation):
        self.observations.append(observation)

    def average(self):
        total = 0.0
        norm = 0.0
        for mass, weight in self.observations:
            total += mass * weight
            norm += weight
        if norm == 0:
            return 0.0
        return total / norm

    def test(self, mass, max_delta=0.01):
        return abs(self.mass - mass) < max_delta

    def __repr__(self):
        return "{self.__class__.__name__}({self.mass}, {self.observations})"



"""A collection of methods for determining whether a given spectrum is
of high quality (likely to produce a high quality interpretation)
"""
from .heuristic import xrea
from .isolation import CoIsolation, PrecursorPurityEstimator

__all__ = [
    "xrea",
    "CoIsolation", "PrecursorPurityEstimator"
]



try:
    from ms_deisotope._c.units import patch_pyteomics
    patch_pyteomics()
except ImportError:
    pass

from typing import Any
from .infer_type import MSFileLoader
from .mzml import MzMLLoader
from .mzxml import MzXMLLoader
from .mgf import MGFLoader
from .common import (
    Scan,
    PrecursorInformation, ProcessedScan,
    IsolationWindow,
    ScanDataSource, ScanIterator, ScanBunch,
    RandomAccessScanSource, ChargeNotProvided)

from .metadata.file_information import (
    FileInformation,
    FileFormat, file_formats,
    FileContent, content_keys,
    IDFormat, id_formats)

from .metadata.instrument_components import (
    Component, ComponentGroup, InstrumentInformation,
    components)

from .metadata.scan_traits import (
    ScanAcquisitionInformation, ScanEventInformation,
    ScanWindow, scan_attributes)

from .metadata.activation import (
    ActivationInformation, MultipleActivationInformation,
    DissociationMethod, dissociation_methods)

from .scan import ScanProxyContext

from ._compression import get_opener

from .text import scan_from_csv
from .memory import make_scan, ScanCollection, make_scan_collection


ProcessedRandomAccessScanSource = RandomAccessScanSource[Any, ProcessedScan]


__all__ = [
    "MSFileLoader", "MzMLLoader",
    "MzXMLLoader", "MGFLoader",

    "Scan", "ProcessedScan",

    "PrecursorInformation",

    "ActivationInformation", "MultipleActivationInformation",
    "DissociationMethod", "dissociation_methods",

    "ScanAcquisitionInformation", "ScanEventInformation",
    "IsolationWindow", "ScanWindow", "scan_attributes",

    "FileInformation", "FileFormat", "file_formats",
    "FileContent", "content_keys",
    "IDFormat", "id_formats",

    "Component", "ComponentGroup", "InstrumentInformation",
    "components",

    "ScanDataSource", "ScanIterator", "ScanBunch",
    "ScanWindow", "RandomAccessScanSource", "ChargeNotProvided",
    "get_opener", "ScanProxyContext", 'scan_from_csv',
    "make_scan", "ScanCollection", "make_scan_collection",
    "ProcessedRandomAccessScanSource",
]



from ms_deisotope._c.spectrum_graph import (
    PathFinder,
    MassWrapper,
    PeakGroupNode,
    PeakNode,
    NodeBase,
    Path,
    SpectrumGraph)

amino_acids = [
    MassWrapper('G', 57.02146372057),
    MassWrapper('A', 71.03711378471),
    MassWrapper('S', 87.03202840427),
    MassWrapper('P', 97.05276384884999),
    MassWrapper('V', 99.06841391299),
    MassWrapper('T', 101.04767846841),
    MassWrapper('C', 103.00918478471),
    MassWrapper('J', 113.08406397713),
    MassWrapper('N', 114.04292744114),
    MassWrapper('D', 115.02694302383),
    MassWrapper('Q', 128.05857750528),
    MassWrapper('K', 128.094963014),
    MassWrapper('E', 129.04259308797),
    MassWrapper('M', 131.04048491299),
    MassWrapper('H', 137.05891185845),
    MassWrapper('F', 147.06841391299),
    MassWrapper('R', 156.1011110236),
    MassWrapper('Y', 163.06332853255),
    MassWrapper('W', 186.07931294986),
]

def find_paths(peaks, components=None, error_tolerance=1e-5, merge=False):
    if components is None:
        components = amino_acids
    sequencer = PathFinder(components, error_tolerance)
    paths = sequencer.paths(peaks, merge=merge)
    return paths



from ms_deisotope.utils import Base
from ms_deisotope.feature_map.feature_map import NeutralMassIndex


def ppm_error(x, y):
    return abs(x - y) / y


def binsearch(array, value):
    lo = 0
    hi = len(array)
    while hi != lo:
        mid = (hi + lo) / 2
        point = array[mid]
        if value == point:
            return mid
        elif hi - lo == 1:
            return mid
        elif point > value:
            hi = mid
        else:
            lo = mid


class FeatureRelationshipBase(object):
    def __repr__(self):
        return "{self.__class__.__name__}()".format(self=self)

    def __hash__(self):
        return hash(repr(self))


class ChargeRelationship(FeatureRelationshipBase):
    def __init__(self, charge_diff, error_tolerance=2e-5):
        self.charge_diff = charge_diff
        self.error_tolerance = error_tolerance

    def __repr__(self):
        return "{self.__class__.__name__}(charge_diff={self.charge_diff})".format(self=self)

    def test(self, a, b):
        if ppm_error(a.neutral_mass, b.neutral_mass) < self.error_tolerance:
            if (a.charge - b.charge) == self.charge_diff:
                return True
        else:
            return False

    def find(self, feature, index):
        hits = index.find_all(feature.neutral_mass, self.error_tolerance)
        for hit in hits:
            if self.test(feature, hit):
                relation = Relation(feature, hit, self)
                feature.supporters.append(relation)


class MassShiftRelationship(FeatureRelationshipBase):
    def __init__(self, mass_shift, error_tolerance=2e-5):
        self.mass_shift = mass_shift
        self.error_tolerance = error_tolerance

    def __repr__(self):
        return "{self.__class__.__name__}(mass_shift={self.mass_shift})".format(self=self)

    def test(self, a, b):
        return ppm_error(
            a.neutral_mass + self.mass_shift, b.neutral_mass) < self.error_tolerance

    def find(self, feature, index):
        hits = index.find_all(feature.neutral_mass + self.mass_shift, self.error_tolerance)
        for hit in hits:
            if self.test(feature, hit):
                relation = Relation(feature, hit, self)
                feature.supporters.append(relation)


class FittedRelationshipFunction(object):
    def __init__(self, relationship, coefficient):
        self.relationship = relationship
        self.coefficient = coefficient

    def test(self, a, b):
        return self.relationship.test(a, b)


class Relation(Base):
    def __init__(self, base, reference, relation):
        self.base = base
        self.reference = reference
        self.relation = relation


class FeatureRelationshipFitter(object):
    def __init__(self, feature_relations=None):
        if feature_relations is None:
            feature_relations = []
        self.feature_relations = feature_relations
        self.history = []

    def add_history(self, hist):
        self.history.append(hist)

    def fit(self, features, past_reference=None, minimum_mass=500):
        relations = []
        features = NeutralMassIndex(features)
        if len(features) == 0:
            return relations
        start_ix = binsearch([f.neutral_mass for f in features], minimum_mass)
        for base in features[start_ix:]:
            for feat_rel_rule in self.feature_relations:
                feat_rel_rule.find(base, features)

        if past_reference is not None and len(past_reference) > 0:
            past_reference = NeutralMassIndex(past_reference)
            for base in features[start_ix:]:
                for feat_rel_rule in self.feature_relations:
                    feat_rel_rule.find(base, past_reference)

    def predict(self, features):
        for feature in features:
            feature.score *= 1.0 + (len(feature.supporters) * 0.05)



# pragma: no cover
try:
    from matplotlib import font_manager
    from matplotlib import pyplot as plt
    from matplotlib import patches as mpatches
    from matplotlib.textpath import TextPath
except (RuntimeError, ImportError):
    pass

from ms_peak_picker.utils import draw_peaklist

from .subgraph import (layout_layers, peak_overlap)

font_options = font_manager.FontProperties(family='monospace')


def draw_envelope_subgraph(envelopes, scale_factor=1.0, overlap_fn=peak_overlap, ax=None, **kwargs):
    layers = layout_layers(envelopes)
    max_score = max(e.score for e in envelopes)
    peaks = set()

    for e in envelopes:
        peaks.update(e.fit.experimental)

    peaks = sorted(peaks, key=lambda x: x.mz)
    peaks = [p.clone() for p in peaks]
    total_intensity = sum(p.intensity for p in peaks)

    start = peaks[0].mz
    end = peaks[-1].mz

    if ax is None:
        figure, ax = plt.subplots(1, 1)

    row_width = float('inf')
    annotation_text_size = 3. * scale_factor
    layer_height = 0.56 * scale_factor
    y_step = (layer_height + 0.05) * - scale_factor
    origin_y = cur_y = -layer_height - 0.075

    cur_position = peaks[0].mz

    for layer in layers:
        layer.sort(key=lambda x: x.start)

    while cur_position < end:
        next_row = cur_position + row_width

        for layer in layers:
            c = 0
            for envelope in layer:
                if envelope.start < cur_position:
                    continue
                elif envelope.start > next_row:
                    break
                c += 1
                rect = mpatches.Rectangle(
                    (envelope.start - 0.01, cur_y),
                    width=0.01 + envelope.end - envelope.start, height=layer_height,
                    facecolor='lightblue', edgecolor='black', linewidth=0.15,
                    alpha=min(max(envelope.score / max_score, 0.2), 0.8))
                ax.add_patch(rect)
                text_path = TextPath(
                    (envelope.start + 0.1, cur_y + 0.2),
                    "%0.2f, %d" % (envelope.score, envelope.fit.charge), size=annotation_text_size / 14.5,
                    prop=font_options, stretch=200)
                patch = mpatches.PathPatch(text_path, facecolor='grey', lw=0.04)
                ax.add_patch(patch)

            if c > 0:
                cur_y += y_step
        cur_y += y_step / 5
        cur_position = next_row

    for p in peaks:
        p.intensity = (p.intensity / total_intensity) * abs(origin_y - cur_y) * 8

    draw_peaklist(peaks, ax=ax)

    ax.set_ylim(cur_y, max(p.intensity for p in peaks) + 0.2)
    ax.set_xlim(start - 0.2, end + 0.2)
    ax.axes.get_yaxis().set_visible(False)
    return ax



# cython: language_level=3
cimport cython

from cpython.object cimport PyObject
from cpython.dict cimport PyDict_GetItem, PyDict_SetItem
from cpython.list cimport PyList_Append


from ms_deisotope._c.compat cimport PyStr_AsUTF8AndSize, PyStr_FromString


cpdef reconstruct_unit_struct(cls, value, unit_info):
    inst = cls(value)
    inst.unit_info = unit_info
    return inst


cpdef clear_unit_cv_table():
    _UNIT_CV_INTERN_TABLE.clear()


cdef str _intern_unit_or_cv(str unit_or_cv):
    """Intern `unit_or_cv` in :const:`~._UNIT_CV_INTERN_TABLE`, potentially
    keeping a reference to the object stored for the duration of the program.
    Parameters
    ----------
    unit_or_cv : object
        The value to intern
    Returns
    -------
    object:
        The object which `unit_or_cv` hash-equals in :const:`~._UNIT_CV_INTERN_TABLE`.
    """
    cdef:
        PyObject* ptmp
    if unit_or_cv is None:
        return None
    ptmp = PyDict_GetItem(_UNIT_CV_INTERN_TABLE, unit_or_cv)
    if ptmp == NULL:
        PyDict_SetItem(_UNIT_CV_INTERN_TABLE, unit_or_cv, unit_or_cv)
        return <str>PyDict_GetItem(_UNIT_CV_INTERN_TABLE, unit_or_cv)
    else:
        return <str>ptmp


@cython.final
@cython.no_gc
cdef class UnitInt(int):
    """Represents an integer value with a unit name.
    Behaves identically to a built-in :class:`int` type.
    Attributes
    ----------
    unit_info : :class:`str`
        The name of the unit this value posseses.
    """
    cdef:
        public str unit_info

    @classmethod
    def create(cls, value, unit_info):
        cdef UnitInt self = UnitInt(value)
        self.unit_info = _intern_unit_or_cv(unit_info)
        return self

    def __reduce__(self):
        return reconstruct_unit_struct, (self.__class__, int(self), self.unit_info)

    def _repr_pretty_(self, p, cycle):
        base = super(UnitInt, self).__repr__()
        if self.unit_info:
            string = "%s %s" % (base, self.unit_info)
        else:
            string = base
        p.text(string)


@cython.final
@cython.no_gc
cdef class UnitFloat(float):
    """Represents an float value with a unit name.
    Behaves identically to a built-in :class:`float` type.
    Attributes
    ----------
    unit_info : :class:`str`
        The name of the unit this value posseses.
    """
    cdef:
        public str unit_info

    @classmethod
    def create(cls, value, unit_info):
        cdef UnitFloat self = UnitFloat(value)
        self.unit_info = _intern_unit_or_cv(unit_info)
        return self

    def __reduce__(self):
        return reconstruct_unit_struct, (self.__class__, float(self), self.unit_info)

    def _repr_pretty_(self, p, cycle):
        base = super(UnitFloat, self).__repr__()
        if self.unit_info:
            string = "%s %s" % (base, self.unit_info)
        else:
            string = base
        p.text(string)


@cython.final
@cython.no_gc
cdef class UnitStr(str):
    """Represents an string value with a unit name.
    Behaves identically to a built-in :class:`str` type.
    Attributes
    ----------
    unit_info : :class:`str`
        The name of the unit this value posseses.
    """
    cdef:
        public str unit_info

    @classmethod
    def create(cls, value, unit_info):
        cdef UnitStr self = UnitStr(value)
        self.unit_info = _intern_unit_or_cv(unit_info)
        return self

    def __reduce__(self):
        return reconstruct_unit_struct, (self.__class__, str(self), self.unit_info)

    def _repr_pretty_(self, p, cycle):
        base = super(UnitStr, self).__repr__()
        if self.unit_info:
            string = "%s %s" % (base, self.unit_info)
        else:
            string = base
        p.text(string)


cpdef reconstruct_cvstr(value, accession, unit_accession):
    return CVStr.create(value, accession, unit_accession)


cdef dict _UNIT_CV_INTERN_TABLE = dict()


cdef dict CVStr_cache = {}


@cython.final
@cython.no_gc
cdef class CVStr(str):
    """A helper class to associate a controlled vocabullary accession
    number with an otherwise plain :class:`str` object
    Attributes
    ----------
    accession : str
        The accession number for this parameter, e.g. MS:1000040
    unit_accession : str
        The accession number for the unit of the value, if any
    """
    cdef:
        public str accession
        public str unit_accession

    @classmethod
    def create(cls, value, accession=None, unit_accession=None):
        cdef:
            PyObject* ptmp
            CVStr self
        accession = _intern_unit_or_cv(accession)
        unit_accession = _intern_unit_or_cv(unit_accession)
        ptmp = PyDict_GetItem(CVStr_cache, value)
        if ptmp != NULL:
            self = <CVStr>ptmp
            if self.accession is accession and self.unit_accession is unit_accession:
                return  self

        self = CVStr(value)
        self.accession = accession
        self.unit_accession = unit_accession

        PyDict_SetItem(CVStr_cache, self, self)
        return self

    def __reduce__(self):
        return reconstruct_cvstr, (str(self), self.accession, self.unit_accession)

    def _repr_pretty_(self, p, cycle):
        base = super(CVStr, self).__repr__()
        p.text(base)


cdef class UnitPrimitiveWrapperMeta(type):
    cdef:
        public type wrapped

    def __init__(self, name, bases, state):
        super().__init__(name, bases, state)
        self.wrapped = state['wrapped']

    def __instancecheck__(cls, inst):
        return isinstance(inst, cls.wrapped)

    def __call__(cls, value, unit_info=None):
        return cls.wrapped.create(value, unit_info)

    def __eq__(self, other):
        if isinstance(other, UnitPrimitiveWrapperMeta):
            return self.wrapped == other.wrapped
        else:
            return self.wrapped == other


class unitint(metaclass=UnitPrimitiveWrapperMeta):
    wrapped = UnitInt


class unitfloat(metaclass=UnitPrimitiveWrapperMeta):
    wrapped = UnitFloat


class unitstr(metaclass=UnitPrimitiveWrapperMeta):
    wrapped = UnitStr


cdef class CVPrimitiveWrapperMeta(type):
    cdef:
        public type wrapped

    def __init__(self, name, bases, state):
        super().__init__(name, bases, state)
        self.wrapped = state['wrapped']

    def __instancecheck__(cls, inst):
        return isinstance(inst, cls.wrapped)

    def __call__(cls, value, accession=None, unit_accession=None):
        return cls.wrapped.create(value, accession, unit_accession)

    def __eq__(self, other):
        if isinstance(other, UnitPrimitiveWrapperMeta):
            return self.wrapped == other.wrapped
        else:
            return self.wrapped == other

    def __hash__(self):
        return hash(self.__qualname__)

class cvstr(metaclass=CVPrimitiveWrapperMeta):
    wrapped = CVStr


@cython.freelist(10)
cdef class _XMLParam:
    cdef:
        public CVStr name
        public object value
        public str type

    cpdef bint is_empty(self):
        value = self.value
        return value == "" or value is None

    def __iter__(self):
        yield self.name
        yield self.value
        yield self.type

    def __init__(self, name, value, type):
        self.name = name
        self.value = value
        self.type = type

    @staticmethod
    cdef _XMLParam _create(CVStr name, object value, str type):
        cdef _XMLParam self = _XMLParam.__new__(_XMLParam)
        self.name = name
        self.value = value
        self.type = type
        return self

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name}, {self.value}, {self.type})"


cpdef str _local_name(element):
    cdef:
        Py_ssize_t i, n
        char* cname
        str tag, result

    tag = element.tag
    if tag and tag[0] == '{':
        cname = PyStr_AsUTF8AndSize(tag, &n)
        for i in range(n - 1, 0, -1):
            if cname[i] == 125: # '}'
                result = PyStr_FromString(&cname[i + 1])
                return result
        return tag
    else:
        return tag


@cython.binding(True)
def _handle_param(self, element, **kwargs):
    """Unpacks cvParam and userParam tags into key-value pairs"""
    cdef:
        object value, attribs
        str unit_info, unit_accession, accession, param_type

    attribs = element.attrib
    unit_info = None
    unit_accesssion = None
    if 'unitCvRef' in attribs or 'unitName' in attribs:
        unit_accesssion = attribs.get('unitAccession')
        unit_name = attribs.get('unitName', unit_accesssion)
        unit_info = unit_name

    accession = attribs.get('accession')
    value = attribs.get('value', '')
    param_type = attribs.get('type')

    if param_type is not None:
        if param_type == 'xsd:int':
            value = UnitInt.create(value, unit_info)
        elif param_type == 'xsd:float':
            value = UnitFloat.create(value, unit_info)
        elif param_type == 'xsd:string':
            value = UnitStr.create(value, unit_info)
        else:
            try:
                if value and value[0].isdigit():
                    value = UnitFloat.create(value, unit_info)
                else:
                    value = UnitStr.create(value, unit_info)
            except ValueError:
                value = UnitStr.create(value, unit_info)
    else:
        try:
            if value and value[0].isdigit():
                value = UnitFloat.create(value, unit_info)
            else:
                value = UnitStr.create(value, unit_info)
        except ValueError:
            value = UnitStr.create(value, unit_info)

    return _XMLParam._create(CVStr.create(attribs['name'], accession, unit_accesssion), value, _local_name(element))


@cython.binding(True)
def _handle_param_fill_missing_value(self, element, **kwargs):
    attribs = element.attrib
    if "value" not in attribs:
        attribs['value'] = ''
    return _handle_param(self, element)


def patch_pyteomics():
    # Cannot patch here or else we can introduce issues
    # with pickling.
    # from pyteomics.auxiliary import structures
    # structures.unitint = unitint
    # structures.unitfloat = unitfloat
    # structures.unitstr = unitstr
    # structures.cvstr = cvstr
    from pyteomics import xml
    xml.unitint = unitint
    xml.unitfloat = unitfloat
    xml.unitstr = unitstr
    xml.cvstr = cvstr
    xml.XML._handle_param = _handle_param
    xml._local_name = _local_name


from .metadata.instrument_components import (
    Component, component, all_components,
    ComponentGroup, InstrumentInformation)

from .metadata.file_information import (
    FileInformation, SourceFile)

from .metadata.scan_traits import (
    IsolationWindow,
    ScanAcquisitionInformation,
    ScanEventInformation,
    ScanWindow)

from .metadata.activation import (
    ActivationInformation, MultipleActivationInformation,
    dissociation_methods_map as dissociation_methods,
    HCD, CID, ETD, ECD, UnknownDissociation)

from .scan import (
    ScanBunch, Scan, ProcessedScan, ScanBase,
    PrecursorInformation, WrappedScan, AveragedScan,
    RawDataArrays, ChargeNotProvided,
    DEFAULT_CHARGE_WHEN_NOT_RESOLVED,
    _ScanIteratorImplBase, _SingleScanIteratorImpl,
    _FakeGroupedScanIteratorImpl, _GroupedScanIteratorImpl,
    _InterleavedGroupedScanIteratorImpl,
    ScanDataSource, ScanIterator, RandomAccessScanSource,
    ScanFileMetadataBase)


__all__ = [
    "Scan", "ScanBunch", "ProcessedScan", "WrappedScan", "ScanBase",
    "AveragedScan", "PrecursorInformation", "RawDataArrays",

    "ScanAcquisitionInformation", "ScanEventInformation", "ScanWindow",
    "IsolationWindow",

    "ScanDataSource", "ScanIterator", "RandomAccessScanSource",
    "ScanFileMetadataBase",

    "_ScanIteratorImplBase", "_SingleScanIteratorImpl",
    "_FakeGroupedScanIteratorImpl", "_GroupedScanIteratorImpl",
    "_InterleavedGroupedScanIteratorImpl",

    "ChargeNotProvided", "DEFAULT_CHARGE_WHEN_NOT_RESOLVED",

    "ActivationInformation", "MultipleActivationInformation", "dissociation_methods",
    "HCD", "CID", "ETD", "ECD", "UnknownDissociation",

    "FileInformation", "SourceFile",

    "Component", "component", "all_components", "ComponentGroup",
    "InstrumentInformation",
]




import logging
import os
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
import warnings

from six import string_types as basestring

import numpy as np

from ms_deisotope import constants, MSFileLoader
from ms_deisotope.data_source.scan.base import ScanBunch
from ms_deisotope.data_source.scan.loader import ScanIterator
from ms_deisotope.task import LogUtilsMixin
from ms_deisotope.utils import Base
from ms_deisotope.averagine import AveragineCache, PROTON

from ms_deisotope.data_source.scan.mobility_frame import IonMobilityFrame, IonMobilitySourceRandomAccessFrameSource, Generic3DIonMobilityFrameSource

logger = logging.getLogger("ms_deisotope.frame_processor")
logger.addHandler(logging.NullHandler())


def _loader_creator(specification, **kwargs):
    if isinstance(specification, (tuple, list)):
        specification, options = specification
        if options:
            kwargs.update(options)
    if isinstance(specification, (basestring, os.PathLike)):
        return MSFileLoader(specification, **kwargs)
    if isinstance(specification, IonMobilitySourceRandomAccessFrameSource):
        return specification
    elif isinstance(specification, ScanIterator):
        return Generic3DIonMobilityFrameSource(specification, **kwargs)
    else:
        raise ValueError(
            "Cannot determine how to get a ScanIterator from %r" % (specification,))


class IonMobilityFrameProcessor(Base, LogUtilsMixin):
    data_source: Union[os.PathLike, IonMobilitySourceRandomAccessFrameSource]
    loader_type: Callable

    ms1_averaging: int
    msn_averaging: int
    num_threads: int

    ms1_peak_picking_args: Dict[str, Any]
    msn_peak_picking_args: Dict[str, Any]

    ms1_deconvolution_args: Dict[str, Any]
    msn_deconvolution_args: Dict[str, Any]


    def __init__(self, data_source, ms1_peak_picking_args=None,
                 msn_peak_picking_args=None,
                 ms1_deconvolution_args=None,
                 msn_deconvolution_args=None,
                 loader_type=None,
                 terminate_on_error=True,
                 ms1_averaging=0,
                 msn_averaging=0, num_threads=3):
        if loader_type is None:
            loader_type = _loader_creator

        self.data_source = data_source
        self.ms1_peak_picking_args = ms1_peak_picking_args or {}
        self.msn_peak_picking_args = msn_peak_picking_args or ms1_peak_picking_args or {}
        self.ms1_deconvolution_args = ms1_deconvolution_args or {}
        self.ms1_deconvolution_args.setdefault("charge_range", (1, 8))
        self.msn_deconvolution_args = msn_deconvolution_args or {}
        self.msn_deconvolution_args.setdefault("charge_range", (1, 8))

        self.ms1_averaging = int(ms1_averaging) if ms1_averaging else 0
        self.msn_averaging = int(msn_averaging) if msn_averaging else 0
        self.num_threads = num_threads
        self.loader_type = loader_type

        self._signal_source = self.loader_type(data_source)
        self.terminate_on_error = terminate_on_error
        self._prepopulate_averagine_cache()

    @property
    def reader(self) -> IonMobilitySourceRandomAccessFrameSource:
        """The :class:`~.IonMobilitySourceRandomAccessFrameSource` which generates the raw scans that will
        be processed.

        Returns
        -------
        :class:`~.IonMobilitySourceRandomAccessFrameSource`
        """
        return self._signal_source

    def _prepopulate_averagine_cache(self):
        if 'averagine' in self.ms1_deconvolution_args:
            averagine = self.ms1_deconvolution_args['averagine']
            ms1_truncate_after = self.ms1_deconvolution_args.get(
                'truncate_after', constants.TRUNCATE_AFTER)
            ms1_ignore_below = self.ms1_deconvolution_args.get(
                'ignore_below', constants.IGNORE_BELOW)
            ms1_charge_range = self.ms1_deconvolution_args.get(
                'charge_range', (1, 8))
            ms1_charge_carrier = self.ms1_deconvolution_args.get(
                'charge_carrier', PROTON)
            if isinstance(averagine, (list, tuple)):
                averagine = [
                    AveragineCache(a).populate(
                        truncate_after=ms1_truncate_after,
                        ignore_below=ms1_ignore_below,
                        min_charge=ms1_charge_range[0],
                        max_charge=ms1_charge_range[1],
                        charge_carrier=ms1_charge_carrier)
                    for a in averagine]
            else:
                averagine = AveragineCache(averagine).populate(
                    truncate_after=ms1_truncate_after,
                    ignore_below=ms1_ignore_below,
                    min_charge=ms1_charge_range[0],
                    max_charge=ms1_charge_range[1],
                    charge_carrier=ms1_charge_carrier)
            self.ms1_deconvolution_args['averagine'] = averagine
        if 'averagine' in self.msn_deconvolution_args:
            averagine = self.msn_deconvolution_args['averagine']
            msn_truncate_after = self.msn_deconvolution_args.get(
                'truncate_after', constants.TRUNCATE_AFTER)
            msn_ignore_below = self.msn_deconvolution_args.get(
                'ignore_below', constants.IGNORE_BELOW)
            msn_charge_range = self.msn_deconvolution_args.get(
                'charge_range', (1, 8))
            msn_charge_carrier = self.msn_deconvolution_args.get(
                'charge_carrier', PROTON)
            if isinstance(averagine, (list, tuple)):
                averagine = [
                    AveragineCache(a).populate(
                        truncate_after=msn_truncate_after,
                        ignore_below=msn_ignore_below,
                        min_charge=msn_charge_range[0],
                        max_charge=msn_charge_range[1],
                        charge_carrier=msn_charge_carrier
                    ) for a in averagine]
            else:
                averagine = AveragineCache(averagine).populate(
                    truncate_after=msn_truncate_after,
                    ignore_below=msn_ignore_below,
                    min_charge=msn_charge_range[0],
                    max_charge=msn_charge_range[1],
                    charge_carrier=msn_charge_carrier)
            self.msn_deconvolution_args['averagine'] = averagine

    def process_frame_group(self, precursor_frame: IonMobilityFrame, product_frames: List[IonMobilityFrame]) -> Tuple[IonMobilityFrame, None, List[IonMobilityFrame]]:
        if precursor_frame is not None:
            self.extract_precursor_features(precursor_frame)
        return precursor_frame, None, product_frames

    def _default_max_gap_size(self, frame: IonMobilityFrame) -> float:
        delta_im = np.median(np.diff(frame.ion_mobilities))
        max_gap_size = (delta_im * 3) + (delta_im / 2)
        return max_gap_size

    def extract_precursor_features(self, precursor_frame: IonMobilityFrame):
        self.log(f"Extracting MS1 raw features from {precursor_frame.id}")
        options = self.ms1_peak_picking_args.copy()
        options.setdefault("average_within", self.ms1_averaging)
        options.setdefault("num_threads", self.num_threads)

        if "max_gap_size" not in options:
            options["max_gap_size"] = self._default_max_gap_size(precursor_frame)

        precursor_frame.extract_features(**options)

    def deconvolute_precursor_features(self, precursor_frame: IonMobilityFrame):
        self.log(f"Deconvolving MS1 IMS features from {precursor_frame.id}")
        options = self.ms1_deconvolution_args.copy()

        if "max_gap_size" not in options:
            options["max_gap_size"] = self._default_max_gap_size(
                precursor_frame)

        precursor_frame.deconvolute_features(**options)

    def extract_product_features(self, product_frame: IonMobilityFrame):
        self.log(f"Extracting MSn raw features from {product_frame.id}")
        options = self.msn_peak_picking_args.copy()
        options.setdefault("average_within", self.msn_averaging)
        options.setdefault("num_threads", self.num_threads)

        if "max_gap_size" not in options:
            options["max_gap_size"] = self._default_max_gap_size(
                product_frame)

        product_frame.extract_features(**options)

    def deconvolute_product_features(self, product_frame: IonMobilityFrame):
        self.log(f"Deconvolving MSn IMS features from {product_frame.id}")
        options = self.msn_deconvolution_args.copy()

        if "max_gap_size" not in options:
            options["max_gap_size"] = self._default_max_gap_size(
                product_frame)

        product_frame.deconvolute_features(**options)

    def process(self, precursor: IonMobilityFrame, products: List[IonMobilityFrame]) -> ScanBunch:
        precursor, _, products = self.process_frame_group(precursor, products)
        if precursor is not None:
            self.deconvolute_precursor_features(precursor)
        for product in products:
            self.extract_product_features(product)
            self.deconvolute_product_features(product)
        return ScanBunch(precursor, products)

    def __next__(self):
        precursor, products = next(self.reader)
        return self.process(precursor, products)

    def __iter__(self) -> 'IonMobilityFrameProcessor':
        return self

    def start_from_scan(self, *args, **kwargs) -> 'IonMobilityFrameProcessor':
        """A wrapper around :meth:`~.IonMobilitySourceRandomAccessFrameSource.start_from_scan` provided by
        :attr:`reader`, if available.

        Returns
        -------
        self

        See Also
        --------
        :meth:`~.IonMobilitySourceRandomAccessFrameSource.start_from_scan`
        """
        self.reader.start_from_scan(*args, **kwargs)
        return self


IonMobilityFrameProcessor.log_with_logger(logger)


# cython: embedsignature =True

cimport cython

from ms_peak_picker._c.peak_set cimport PeakBase
from ms_deisotope._c.feature_map.lcms_feature cimport LCMSFeature, LCMSFeatureTreeNode
from cpython.list cimport PyList_GET_SIZE, PyList_GET_ITEM, PyList_GetSlice, PyList_GetItem

cimport numpy as np
import numpy as np

np.import_array()

from cpython cimport array
import array



cdef class LCMSFeatureMap(object):
    def __init__(self, features):
        self.features = sorted(features, key=lambda x: x.mz)

    def __len__(self):
        return PyList_GET_SIZE(self.features)

    def __iter__(self):
        return iter(self.features)

    def __getitem__(self, i):
        if isinstance(i, (int, slice)):
            return self.features[i]
        else:
            return [self.features[j] for j in i]

    def __eq__(self, other):
        if other is None:
            return False
        return self.features == other.features

    def __ne__(self, other):
        return not self == other

    cdef Py_ssize_t get_size(self):
        return PyList_GET_SIZE(self.features)

    cdef LCMSFeature get(self, size_t i):
        return <LCMSFeature>PyList_GetItem(self.features, i)

    cpdef LCMSFeature search(self, double mz, double error_tolerance=2e-5):
        """Search for a single feature within `error_tolerance` of `mz`.

        Parameters
        ----------
        mz : float
            The m/z to search for
        error_tolerance : float, optional
            The error tolerance to search with, defaults to 2e-5, 20 PPM

        Returns
        -------
        :class:`~.LCMSFeature`
            The a feature matching the query m/z, or :const:`None`.

        See Also
        --------
        find_all
        between
        """

        cdef:
            long i
            LCMSFeature out
        i = binary_search(self.features, mz, error_tolerance)
        if i == -1:
            return None
        else:
            out = <LCMSFeature>PyList_GET_ITEM(self.features, i)
            return out

    cpdef list find_all(self, double mz, double error_tolerance=2e-5):
        """Search for all features within `error_tolerance` of `mz`.

        Parameters
        ----------
        mz : float
            The m/z to search for
        error_tolerance : float, optional
            The error tolerance to search with, defaults to 2e-5, 20 PPM

        Returns
        -------
        list of :class:`~.LCMSFeature`
            A list of all features matching the query m/z, or the empty list.

        See Also
        --------
        between
        """
        cdef:
            long loout, hiout
            list result
        search_sweep(self.features, mz, error_tolerance, &loout, &hiout)
        result = <list>PyList_GetSlice(self.features, loout, hiout)
        return result

    def index_range(self, lo, hi, error_tolerance=2e-5):
        return (
            binary_search_with_flag(
                self.features, lo, error_tolerance)[0][0],
            binary_search_with_flag(
                self.features, hi, error_tolerance)[0][0])

    cpdef list spanning_time(self, double time_point):
        cdef:
            size_t i, n
            LCMSFeature feature
            list result
        result = []
        n = self.get_size()
        for i in range(n):
            feature = self.get(i)
            if feature.spans_in_time(time_point):
                result.append(feature)
        return result

    cpdef list between(self, double lo, double hi, double error_tolerance=2e-5):
        """Search for all features between `lo` and `hi`, allowing `error_tolerance`
        around the edges.

        Parameters
        ----------
        lo : float
            The m/z to search for the lower bound
        hi : float
            The m/z to search for the upper bound
        error_tolerance : float, optional
            The error tolerance to search with, defaults to 2e-5, 20 PPM

        Returns
        -------
        list of :class:`~.LCMSFeature`
            The features between the query bounds with the permitted error, or an empty list.

        See Also
        --------
        find_all
        """
        cdef:
            Py_ssize_t n, i, lo_ix, hi_ix
            LCMSFeature f

        n = self.get_size()
        if n == 0:
            return []
        lo_ix = binary_search_with_flag(
            self.features, lo, error_tolerance)[0][0]
        if self.get(lo_ix).get_mz() < lo:
            lo_ix += 1
        if lo_ix > n:
            lo_ix = n - 1
        i = lo_ix
        while i < n:
            f = self.get(i)
            if f.get_mz() > hi:
                break
            i += 1
        hi_ix = i
        return self[lo_ix:hi_ix]

    def __repr__(self):
        return "{self.__class__.__name__}(<{size} features>)".format(self=self, size=len(self))

    cpdef LCMSFeatureMap clone(self, bint deep=True):
        cdef:
            size_t i, n
            list result
        result = []
        n = self.get_size()
        for i in range(n):
            feature = self.get(i).clone(deep)
            result.append(feature)
        return self.__class__(result)

    def as_arrays(self):
        cdef:
            size_t i, n, j, m, k, q
            double time
            LCMSFeature feature
            LCMSFeatureTreeNode node
            PeakBase peak

        mz_array = array.array('d')
        intensity_array = array.array('d')
        ion_mobility_array = array.array('d')
        feature_id_array = array.array('L')
        n = self.get_size()
        for i in range(n):
            feature = <LCMSFeature>self.get(i)
            m = feature.get_size()
            for j in range(m):
                node = <LCMSFeatureTreeNode>feature.getitem(j)
                time = node.time
                q = node.get_members_size()
                for k in range(q):
                    peak = <PeakBase>node.getitem(k)
                    ion_mobility_array.append(time)
                    mz_array.append(peak.mz)
                    intensity_array.append(peak.intensity)
                    feature_id_array.append(i)

        mz_array = np.array(mz_array, copy=False)
        intensity_array = np.array(intensity_array, copy=False)
        ion_mobility_array = np.array(ion_mobility_array, copy=False)
        feature_id_array = np.array(feature_id_array, copy=False)
        mask = np.lexsort(np.stack((ion_mobility_array, mz_array)))
        return (mz_array[mask], intensity_array[mask], ion_mobility_array[mask], feature_id_array[mask])


@cython.binding(True)
cpdef list split_sparse(LCMSFeatureMap self, double delta_rt=1.0, size_t min_size=2):
    cdef:
        list result, chunks
        size_t i, n, j, m
        LCMSFeature feature, chunk
    result = []
    n = self.get_size()
    for i in range(n):
        feature = self.get(i)
        chunks = feature.split_sparse(delta_rt)
        m = PyList_GET_SIZE(chunks)
        for j in range(m):
            chunk = <LCMSFeature>PyList_GET_ITEM(chunks, j)
            if chunk.get_size() >= min_size:
                result.append(chunk)
    return result


cpdef tuple binary_search_with_flag(list array, double mz, double error_tolerance):
    cdef:
        int i, n, lo, hi, mid, low_end, high_end
        double err
        LCMSFeature x
    lo = 0
    n = hi = len(array)
    while hi != lo:
        mid = (hi + lo) // 2
        x = <LCMSFeature>PyList_GET_ITEM(array, mid)
        err = (x.get_mz() - mz) / mz
        if abs(err) <= error_tolerance:
            i = mid - 1
            # Begin Sweep forward
            while i > 0:
                x = array[i]
                err = (x.get_mz() - mz) / mz
                if abs(err) <= error_tolerance:
                    i -= 1
                    continue
                else:
                    break
            low_end = i
            i = mid + 1

            # Begin Sweep backward
            while i < n:
                x = array[i]
                err = (x.get_mz() - mz) / mz
                if abs(err) <= error_tolerance:
                    i += 1
                    continue
                else:
                    break
            high_end = i
            return list(range(low_end, high_end)), True
        elif (hi - lo) == 1:
            return [mid], False
        elif err > 0:
            hi = mid
        elif err < 0:
            lo = mid
    return [0], False


@cython.cdivision(True)
cdef long binary_search(list array, double mz, double error_tolerance):
    cdef:
        int lo, hi, n, mid, i
        double err
        LCMSFeature x
    lo = 0
    n = hi = PyList_GET_SIZE(array)
    while hi != lo:
        mid = (hi + lo) / 2
        x = <LCMSFeature>PyList_GET_ITEM(array, mid)
        err = (x.get_mz() - mz) / mz
        if abs(err) <= error_tolerance:
            best_index = mid
            best_error = abs(err)
            i = mid - 1
            while i > 0:
                x = <LCMSFeature>PyList_GET_ITEM(array, i)
                err = abs((x.get_mz() - mz) / mz)
                if err < best_error:
                    best_error = err
                    best_index = i
                i -= 1

            i = mid + 1
            while i < n:
                x = <LCMSFeature>PyList_GET_ITEM(array, i)
                err = abs((x.get_mz() - mz) / mz)
                if err < best_error:
                    best_error = err
                    best_index = i
                i += 1
            return best_index
        elif (hi - lo) == 1:
            return -1
        elif err > 0:
            hi = mid
        elif err < 0:
            lo = mid
    return 0


@cython.cdivision(True)
cdef void search_sweep(list array, double mz, double error_tolerance, long* loout, long* hiout):
    cdef:
        int i, n, lo, hi, mid, start, end
        double err
        LCMSFeature x
    lo = 0
    n = hi = PyList_GET_SIZE(array)
    loout[0] = 0
    hiout[0] = 0
    while hi != lo:
        mid = (hi + lo) / 2
        x = <LCMSFeature>PyList_GET_ITEM(array, mid)
        err = (x.get_mz() - mz) / mz
        if abs(err) <= error_tolerance:
            i = mid - 1
            while i > 0:
                x = <LCMSFeature>PyList_GET_ITEM(array, i)
                err = abs((x.get_mz() - mz) / mz)
                if err > error_tolerance:
                    break
                i -= 1
            start = i + 1
            i = mid + 1
            while i < n:
                x = <LCMSFeature>PyList_GET_ITEM(array, i)
                err = abs((x.get_mz() - mz) / mz)
                if err > error_tolerance:
                    break
                i += 1
            end = i
            loout[0] = start
            hiout[0] = end
            return
        elif (hi - lo) == 1:
            return
        elif err > 0:
            hi = mid
        elif err < 0:
            lo = mid
    return





"""The main entry point for the `ms-deisotope` command line program.
"""
import os

import click

import ms_peak_picker

from ms_peak_picker.scan_filter import parse as parse_filter

import ms_deisotope
from ms_deisotope import MSFileLoader
from ms_deisotope.data_source import RandomAccessScanSource
from ms_deisotope.task.log_utils import init_logging

from ms_deisotope.tools.utils import processes_option, AveragineParamType, is_debug_mode, register_debug_hook
from ms_deisotope.tools.deisotoper import workflow, output



def configure_iterator(loader, start_time, end_time):
    """Configure `loader` to run between the time ranges specified, or
    if the `loader` does not support random access, the bounds are set to
    the start and end of the iterator.

    This function will also set the iteration mode of `loader` to "grouped".

    Parameters
    ----------
    loader : :class:`ScanDataSource`
        The source to load scans from
    start_time : float
        The start time to load scans from
    end_time : float
        The end time to stop loading scans at

    Returns
    -------
    str: start_scan_id
        The id value of the scan to start from, will be :cosnt:`None` if there is no random access.
    float: start_scan_time
        The actual time associated with the start scan
    str: end_scan_id
        The id value of the scan to stop at, will be :cosnt:`None` if there is no random access.
    float: end_scan_time
        The actual time associated with the end scan
    """
    if isinstance(loader, RandomAccessScanSource):
        last_scan = loader[len(loader) - 1]
        last_time = last_scan.scan_time

        start_scan = loader.get_scan_by_time(start_time)
        if loader.has_ms1_scans():
            start_scan = loader._locate_ms1_scan(start_scan)

        if end_time > last_time:
            end_time = last_time
        end_scan = loader.get_scan_by_time(end_time)
        if loader.has_ms1_scans():
            end_scan = loader._locate_ms1_scan(end_scan)

        start_scan_id = start_scan.id
        end_scan_id = end_scan.id

        start_scan_time = start_scan.scan_time
        end_scan_time = end_scan.scan_time

        loader.reset()
        loader.start_from_scan(start_scan_id, grouped=True)
    else:
        click.secho("The file format provided does not support random"
                    " access, start and end points will be ignored", fg='yellow')
        start_scan_time = 0
        start_scan_id = None

        end_scan_time = float('inf')
        end_scan_id = None
        loader.make_iterator(grouped=True)

    return start_scan_id, start_scan_time, end_scan_id, end_scan_time


def check_if_profile(loader):
    """Check if the spectra in `loader` are stored in profile mode,
    established by querying the next scan bunch.

    Parameters
    ----------
    loader : :class:`ScanIterator`
        The source to load scans from

    Returns
    -------
    :class:`bool`:
        Whether or not the spectra are profile mode.
    """
    first_bunch = next(loader)
    if first_bunch.precursor is not None:
        is_profile = (first_bunch.precursor.is_profile)
    elif first_bunch.products:
        is_profile = (first_bunch.products[0].is_profile)

    if is_profile:
        click.secho("Spectra are profile")
    else:
        click.secho("Spectra are centroided", fg='yellow')
    return is_profile


def determine_output_format(output_file: str):
    base, ext = os.path.splitext(output_file)
    gzip_compress_file = False
    if ext.lower() == '.gz':
        gzip_compress_file = True
        base, ext = os.path.splitext(base)
    if ext.lower() == '.mzml':
        storage_handler_type = output.ThreadedMzMLScanStorageHandler
    elif ext.lower() == '.mgf':
        storage_handler_type = output.ThreadedMGFScanStorageHandler
    elif ext.lower() == '.mzmlb':
        storage_handler_type = output.ThreadedMzMLbScanStorageHandler
    else:
        click.secho("Could not infer output file format. Assuming mzML.", fg='yellow')
        storage_handler_type = output.ThreadedMzMLScanStorageHandler
    return storage_handler_type, os.path.basename(base), gzip_compress_file


@click.command("deisotope",
               short_help=(
                   "Convert raw mass spectra data into deisotoped neutral mass peak lists written to mzML."
                   " Can accept mzML, mzXML, MGF with either profile or centroided scans."),
               context_settings=dict(help_option_names=['-h', '--help']))
@click.argument("ms-file", type=click.Path(exists=True))
@click.argument("outfile-path", type=click.Path(writable=True))
@click.option("-a", "--averagine", default=["peptide"],
              type=AveragineParamType(),
              help=('Averagine model to use for MS1 scans. '
                    'Either a name or formula. May specify multiple times.'),
              multiple=True)
@click.option("-an", "--msn-averagine", default=["peptide"],
              type=AveragineParamType(),
              help=('Averagine model to use for MS^n scans. '
                    'Either a name or formula. May specify multiple times.'),
              multiple=True)
@click.option("-s", "--start-time", type=float, default=0.0,
              help='Scan time to begin processing at in minutes')
@click.option("-e", "--end-time", type=float, default=float('inf'),
              help='Scan time to stop processing at in minutes')
@click.option("-c", "--maximum-charge", type=int, default=8,
              help=('Highest absolute charge state to consider'))
@click.option("-n", "--name", default=None,
              help="Name for the sample run to be stored. Defaults to the base name of the input data file")
@click.option("-t", "--score-threshold", type=float, default=workflow.SampleConsumer.MS1_SCORE_THRESHOLD,
              help="Minimum score to accept an isotopic pattern fit in an MS1 scan")
@click.option("-tn", "--msn-score-threshold", type=float, default=workflow.SampleConsumer.MSN_SCORE_THRESHOLD,
              help="Minimum score to accept an isotopic pattern fit in an MS^n scan")
@click.option("-m", "--missed-peaks", type=int, default=3,
              help="Number of missing peaks to permit before an isotopic fit is discarded")
@click.option("-mn", "--msn-missed-peaks", type=int, default=1,
              help="Number of missing peaks to permit before an isotopic fit is discarded in an MSn scan")
@processes_option
@click.option("-b", "--background-reduction", type=float, default=0., help=(
    "Background reduction factor. Larger values more aggresively remove low abundance"
    " signal in MS1 scans."))
@click.option("-bn", "--msn-background-reduction", type=float, default=0., help=(
    "Background reduction factor. Larger values more aggresively remove low abundance"
    " signal in MS^n scans."))
@click.option("-r", '--transform', multiple=True, type=parse_filter, help=(
    "Scan transformations to apply to MS1 scans. May specify more than once."))
@click.option("-rn", '--msn-transform', multiple=True, type=parse_filter, help=(
    "Scan transformations to apply to MS^n scans. May specify more than once."))
@click.option("-v", "--extract-only-tandem-envelopes", is_flag=True, default=False,
              help='Only work on regions that will be chosen for MS/MS')
@click.option("--verbose", is_flag=True, help="Log additional diagnostic information for each scan.")
@click.option("-g", "--ms1-averaging", default=0, type=int, help=(
    "The number of MS1 scans before and after the current MS1 "
    "scan to average when picking peaks."))
@click.option("--ignore-msn", is_flag=True, default=False, help="Ignore MS^n scans")
@click.option("-i", "--isotopic-strictness", default=2.0, type=float)
@click.option("-in", "--msn-isotopic-strictness", default=0.0, type=float)
@click.option("-snr", "--signal-to-noise-threshold", default=1.0, type=float, help=(
    "Signal-to-noise ratio threshold to apply when filtering peaks"))
@click.option("-mo", "--mass-offset", default=0.0, type=float, help=("Shift peak masses by the given amount"))
@click.option("-D", "--default-precursor-ion-selection-window", default=1.5, type=float,
              help="The isolation window width to assume when it is not specified.")
def deisotope(ms_file, outfile_path, averagine=None, start_time=None, end_time=None, maximum_charge=None,
              name=None, msn_averagine=None, score_threshold=35., msn_score_threshold=10., missed_peaks=1,
              msn_missed_peaks=1, background_reduction=0., msn_background_reduction=0.,
              transform=None, msn_transform=None, processes=4, extract_only_tandem_envelopes=False,
              ignore_msn=False, isotopic_strictness=2.0, ms1_averaging=0,
              msn_isotopic_strictness=0.0, signal_to_noise_threshold=1.0, mass_offset=0.0,
              default_precursor_ion_selection_window=1.5, deconvolute=True, verbose=False):
    """Convert raw mass spectra data into deisotoped neutral mass peak lists written to mzML.
    """
    init_logging()
    if transform is None:
        transform = []
    if msn_transform is None:
        msn_transform = []

    if (ignore_msn and extract_only_tandem_envelopes):
        click.secho(
            "Cannot use both --ignore-msn and --extract-only-tandem-envelopes",
            fg='red')
        raise click.Abort("Cannot use both --ignore-msn and --extract-only-tandem-envelopes")

    cache_handler_type, _output_basename, _should_gzip_compress_output_file  = determine_output_format(outfile_path)
    click.echo("Preprocessing %s" % ms_file)
    minimum_charge = 1 if maximum_charge > 0 else -1
    charge_range = (minimum_charge, maximum_charge)

    loader = MSFileLoader(ms_file)
    (start_scan_id, start_scan_time,
     end_scan_id, end_scan_time) = configure_iterator(loader, start_time, end_time)

    is_profile = check_if_profile(loader)

    if loader.has_ms1_scans() == False:
        extract_only_tandem_envelopes = False

    if name is None:
        name = os.path.splitext(os.path.basename(ms_file))[0]

    if os.path.exists(outfile_path) and not os.access(outfile_path, os.W_OK):
        click.secho("Can't write to output file path", fg='red')
        raise click.Abort()

    click.secho("Initializing %s" % name, fg='green')
    click.echo("from %s (%0.2f) to %s (%0.2f)" % (
        start_scan_id, start_scan_time, end_scan_id, end_scan_time))
    if deconvolute:
        click.echo("charge range: %s" % (charge_range,))

    if is_profile:
        ms1_peak_picking_args = {
            "transforms": [
            ] + list(transform),
            "signal_to_noise_threshold": signal_to_noise_threshold
        }
        if background_reduction:
            ms1_peak_picking_args['transforms'].append(
                ms_peak_picker.scan_filter.FTICRBaselineRemoval(
                    scale=background_reduction, window_length=2))
            ms1_peak_picking_args['transforms'].append(ms_peak_picker.scan_filter.SavitskyGolayFilter())
    else:
        ms1_peak_picking_args = {
            "transforms": [
                ms_peak_picker.scan_filter.FTICRBaselineRemoval(
                    scale=background_reduction, window_length=2),
            ] + list(transform)
        }

    if msn_background_reduction > 0.0:
        msn_peak_picking_args = {
            "transforms": [
                ms_peak_picker.scan_filter.FTICRBaselineRemoval(
                    scale=msn_background_reduction, window_length=2),
            ] + list(msn_transform)
        }
    else:
        msn_peak_picking_args = {
            "transforms": [
            ] + list(msn_transform)
        }

    if mass_offset != 0.0:
        ms1_peak_picking_args['transforms'].append(
            ms_peak_picker.scan_filter.RecalibrateMass(offset=mass_offset))
        msn_peak_picking_args['transforms'].append(
            ms_peak_picker.scan_filter.RecalibrateMass(offset=mass_offset))

    if deconvolute:
        if len(averagine) == 1:
            averagine = averagine[0]
            ms1_deconvoluter_type = ms_deisotope.deconvolution.AveraginePeakDependenceGraphDeconvoluter
        else:
            ms1_deconvoluter_type = ms_deisotope.deconvolution.MultiAveraginePeakDependenceGraphDeconvoluter

        ms1_deconvolution_args = {
            "scorer": ms_deisotope.scoring.PenalizedMSDeconVFitter(score_threshold, isotopic_strictness),
            "max_missed_peaks": missed_peaks,
            "averagine": averagine,
            "truncate_after": workflow.SampleConsumer.MS1_ISOTOPIC_PATTERN_WIDTH,
            "ignore_below": workflow.SampleConsumer.MS1_IGNORE_BELOW,
            "deconvoluter_type": ms1_deconvoluter_type,
            "use_quick_charge": True
        }

        if msn_isotopic_strictness >= 1:
            msn_isotopic_scorer = ms_deisotope.scoring.PenalizedMSDeconVFitter(
                msn_score_threshold, msn_isotopic_strictness)
        else:
            msn_isotopic_scorer = ms_deisotope.scoring.MSDeconVFitter(msn_score_threshold)

        if len(msn_averagine) == 1:
            msn_averagine = msn_averagine[0]
            msn_deconvoluter_type = ms_deisotope.deconvolution.AveraginePeakDependenceGraphDeconvoluter
        else:
            msn_deconvoluter_type = ms_deisotope.deconvolution.MultiAveraginePeakDependenceGraphDeconvoluter

        msn_deconvolution_args = {
            "scorer": msn_isotopic_scorer,
            "averagine": msn_averagine,
            "max_missed_peaks": msn_missed_peaks,
            "truncate_after": workflow.SampleConsumer.MSN_ISOTOPIC_PATTERN_WIDTH,
            "ignore_below": workflow.SampleConsumer.MSN_IGNORE_BELOW,
            "use_quick_charge": True,
            "deconvoluter_type": msn_deconvoluter_type,
        }
    else:
        ms1_deconvolution_args = None
        msn_deconvolution_args = None

    consumer = workflow.SampleConsumer(
        ms_file,
        ms1_peak_picking_args=ms1_peak_picking_args,
        ms1_deconvolution_args=ms1_deconvolution_args,
        msn_peak_picking_args=msn_peak_picking_args,
        msn_deconvolution_args=msn_deconvolution_args,
        storage_path=outfile_path, sample_name=name,
        start_scan_id=start_scan_id, storage_type=cache_handler_type,
        end_scan_id=end_scan_id, n_processes=processes,
        extract_only_tandem_envelopes=extract_only_tandem_envelopes,
        ignore_tandem_scans=ignore_msn,
        ms1_averaging=ms1_averaging,
        default_precursor_ion_selection_window=default_precursor_ion_selection_window,
        deconvolute=deconvolute,
        verbose=verbose,
        end_scan_time=end_scan_time,
        start_scan_time=start_scan_time)
    consumer.start()


if is_debug_mode():
    register_debug_hook()

if __name__ == '__main__':
    deisotope.main()
